#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from .backwash_solids_handling_zo import BackwashSolidsHandlingZO
from .bio_active_filtration_zo import BioActiveFiltrationZO
from .blending_reservoir_zo import BlendingReservoirZO
from .conventional_activated_sludge_zo import CASZO
from .cooling_tower_zo import CoolingTowerZO
from .deep_well_injection_zo import DeepWellInjectionZO
from .dmbr_zo import DMBRZO
from .electrodialysis_reversal_zo import ElectrodialysisReversalZO
from .aeration_basin_zo import AerationBasinZO
from .air_flotation_zo import AirFlotationZO
from .bioreactor_zo import BioreactorZO
from .buffer_tank_zo import BufferTankZO
from .chemical_addition_zo import ChemicalAdditionZO
from .cartridge_filtration_zo import CartridgeFiltrationZO
from .cooling_supply_zo import CoolingSupplyZO
from .decarbonator_zo import DecarbonatorZO
from .dissolved_air_flotation_zo import DissolvedAirFlotationZO
from .feed_water_tank_zo import FeedWaterTankZO
from .feed_zo import FeedZO
from .electrochemical_nutrient_removal_zo import ElectroNPZO
from .energy_recovery_zo import EnergyRecoveryZO
from .iron_and_manganese_removal_zo import IronManganeseRemovalZO
from .landfill_zo import LandfillZO
from .pump_electricity_zo import PumpElectricityZO
from .mabr_zo import MABRZO
from .mbr_zo import MBRZO
from .media_filtration_zo import MediaFiltrationZO
from .microfiltration_zo import MicroFiltrationZO
from .microscreen_filtration_zo import MicroscreenFiltrationZO
from .municipal_wwtp_zo import MunicipalWWTPZO
from .nanofiltration_zo import NanofiltrationZO
from .clarifier_zo import ClarifierZO
from .primary_separator_zo import PrimarySeparatorZO
from .pump_zo import PumpZO
from .screen_zo import ScreenZO
from .sedimentation_zo import SedimentationZO
from .settling_pond_zo import SettlingPondZO
from .smp_zo import SMPZO
from .tramp_oil_tank_zo import TrampOilTankZO
from .tri_media_filtration_zo import TriMediaFiltrationZO
from .coag_and_floc_zo import CoagulationFlocculationZO
from .ultra_filtration_zo import UltraFiltrationZO
from .uv_zo import UVZO
from .uv_aop_zo import UVAOPZO
from .walnut_shell_filter_zo import WalnutShellFilterZO
from .waiv_zo import WAIVZO
from .anaerobic_digestion_oxidation_zo import AnaerobicDigestionOxidationZO
from .fixed_bed_zo import FixedBedZO
from .municipal_drinking_zo import MunicipalDrinkingZO
from .ozone_zo import OzoneZO
from .ozone_aop_zo import OzoneAOPZO
from .ion_exchange_zo import IonExchangeZO
from .brine_concentrator_zo import BrineConcentratorZO
from .gac_zo import GACZO
from .dual_media_filtration_zo import DualMediaFiltrationZO
from .storage_tank_zo import StorageTankZO
from .sw_onshore_intake_zo import SWOnshoreIntakeZO
from .chlorination_zo import ChlorinationZO
from .static_mixer_zo import StaticMixerZO
from .co2_addition_zo import CO2AdditionZO
from .sludge_tank_zo import SludgeTankZO
from .evaporation_pond_zo import EvaporationPondZO
from .filter_press_zo import FilterPressZO
from .water_pumping_station_zo import WaterPumpingStationZO
from .well_field_zo import WellFieldZO
from .intrusion_mitigation_zo import IntrusionMitigationZO
from .injection_well_disposal_zo import InjectionWellDisposalZO
from .surface_discharge_zo import SurfaceDischargeZO
from .electrocoagulation_zo import ElectrocoagulationZO


# =========================================================================================
# Units explicitly for Wastewater Resource Recovery effort

from .anaerobic_digestion_reactive_zo import AnaerobicDigestionReactiveZO
from .anaerobic_mbr_mec_zo import AnaerobicMBRMECZO
from .autothermal_hydrothermal_liquefaction_zo import ATHTLZO
from .CANDOP_zo import CANDOPZO
from .centrifuge_zo import CentrifugeZO
from .cloth_media_filtration_zo import ClothMediaFiltrationZO
from .cofermentation_zo import CofermentationZO
from .constructed_wetlands_zo import ConstructedWetlandsZO
from .gas_sparged_membrane_zo import GasSpargedMembraneZO
from .hrcs_zo import HRCSZO
from .hydrothermal_gasification_zo import HTGZO
from .magprex_zo import MagprexZO
from .membrane_evaporator_zo import MembraneEvaporatorZO
from .metab_zo import MetabZO
from .microbial_battery_zo import MicrobialBatteryZO
from .peracetic_acid_disinfection_zo import PeraceticAcidDisinfectionZO
from .photothermal_membrane_zo import PhotothermalMembraneZO
from .secondary_treatment_wwtp_zo import SecondaryTreatmentWWTPZO
from .struvite_classifier_zo import StruviteClassifierZO
from .suboxic_activated_sludge_process_zo import SuboxicASMZO
from .supercritical_salt_precipitation_zo import SaltPrecipitationZO
from .vfa_recovery_zo import VFARecoveryZO



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
#################################################################################

"""
This file contains methods to convert WaterTAP naming conventions to OLI
and generate molecular weight and charge dictionaries from molecular formulae. 

It calculates molecular weights using the periodic_table.csv from:
https://gist.github.com/GoodmanSciences/c2dd862cd38f21b0ad36b8f96b4bf1ee.
"""

__author__ = "Paul Vecchiarelli, Ben Knueven, Adam Atia"

from collections import namedtuple
from re import findall
from pathlib import Path
from os.path import join
from pandas import read_csv
from pyomo.environ import units as pyunits

# TODO: maybe replace some functionality with molmass: https://pypi.org/project/molmass

OLIName = namedtuple(
    "OLIName", ["oli_name", "watertap_name", "charge", "charge_group", "molar_mass"]
)


def watertap_to_oli(watertap_name: str) -> OLIName:
    """
    This method creates a named tuple
    which can be passed directly into OLI
    or into MCAS property models.

    :param watertap_name: string name of substance in WaterTAP format, i.e., B[OH]4_-

    :return OLIName: named tuple containing attributes derived from molecular formula
    """

    c = findall(r"[A-Z]", watertap_name)
    if len(c) == 0:
        raise IOError(
            f" At least 1 uppercase letter is required to specify a molecule, not '{watertap_name}'."
        )

    oli_name = get_oli_name(watertap_name)
    charge = get_charge(watertap_name)
    charge_group = get_charge_group(charge)
    molar_mass = get_molar_mass(watertap_name)
    return OLIName(oli_name, watertap_name, charge, charge_group, molar_mass)


# TODO: merge with other helper functions in this file
def get_oli_name(watertap_name: str) -> str:
    """
    Converts an WaterTAP formatted name, i.e., "Na_+"
    into an OLI formatted name, i.e., "NAION".

    :param watertap_name: string name of a solute in WaterTAP format

    :return oli_name: string name of a solute in OLI format
    """

    components = watertap_name.split("_")
    if len(components) == 0:
        raise IOError(f" Unable to parse solute '{watertap_name}'.")
    if len(components) == 1:
        molecule = components[0]
    elif len(components) == 2:
        molecule = components[0] + "ION"
    oli_name = molecule.replace("[", "").replace("]", "").upper()
    return oli_name


def get_charge(watertap_name: str) -> int:
    """
    Gets charge from WaterTAP formatted names.

    :param watertap_name: string name of a solute in WaterTAP format

    :return charge: integer value of charge
    """

    components = watertap_name.split("_")
    if len(components) == 0:
        raise IOError(f" Unable to parse solute '{watertap_name}'.")
    if len(components) == 1:
        molecule = components[0]
        charge = 0
    elif len(components) == 2:
        molecule = components[0] + "ION"
        charge = components[1]
        try:
            charge_sign = charge[-1]
        except IndexError:
            raise IOError(
                f"Charge sign could not be determined from the string '{watertap_name}'"
            )
        if len(charge) > 1:
            try:
                charge_magnitude = int(charge[:-1])
            except ValueError:
                raise IOError(
                    f"Charge sign could not be determined from the string '{watertap_name}'"
                )
        else:
            charge_magnitude = 1
        if charge_sign == "+":
            charge = charge_magnitude
        elif charge_sign == "-":
            charge = -charge_magnitude
        else:
            raise IOError(
                f"Only + and - are valid charge indicators and neither was provided in '{watertap_name}'."
            )
    else:
        raise IOError(
            f"Charge could not be determined from the string '{watertap_name}'"
        )
    return charge


def get_charge_group(charge: int) -> str:
    """
    Categorizes molecule based on its charge.

    :param charge: integer value for charge

    :return group: string name for charge group
    """

    if charge == 0:
        group = "Neutrals"
    elif charge > 0:
        group = "Cations"
    elif charge < 0:
        group = "Anions"
    return group


def get_molar_mass(watertap_name: str) -> float:
    """
    Extracts atomic weight data from a periodic table file
    to generate the molar mass of a chemical substance.

    TODO: additional testing for complex solutes
    such as CH3CO2H, [UO2]2[OH]4, etc.

    :param watertap_name: string name of a solute in WaterTAP format

    :return molar_mass: float value for molar mass of solute
    """

    file_path = Path(__file__).parents[0]
    periodic_table = read_csv(join(file_path, "periodic_table.csv"))

    components = watertap_name.split("_")
    elements = findall("[A-Z][a-z]?[0-9]*", components[0])
    element_counts = {}
    for element in elements:
        if len(element) == 1:
            element_counts[element] = 1
        elif len(element) == 2 and element.isalpha():
            element_counts[element] = 1
        elif len(element) == 2 and not element.isalpha():
            element_counts[element[:-1]] = int(element[-1])
        elif len(element) == 3 and element[:-1].isalpha():
            element_counts[element[:-1]] = int(element[-1])
        elif len(element) == 3 and not element[:-1].isalpha():
            element_counts[element[:-2]] = int(element[-2:-1])
        else:
            raise IOError(f" Too many characters in {element}.")

        element_location = components[0].find(element)

        if "[" in components[0]:
            boundary = (components[0].find("["), components[0].find("]"))
            coefficient = int(components[0][boundary[1] + 1])
            if element_location > boundary[0] and element_location < boundary[1]:
                element_counts[element] *= coefficient

    molar_mass = 0
    for element in element_counts:
        atomic_mass = float(
            periodic_table["AtomicMass"][(periodic_table["Symbol"] == element)].values[
                0
            ]
        )
        molar_mass += element_counts[element] * atomic_mass

    if not molar_mass:
        raise IOError(f"Molecular weight data could not be found for {watertap_name}.")

    return molar_mass


def get_molar_mass_quantity(watertap_name: str, units=pyunits.kg / pyunits.mol):
    """
    Extracts atomic weight data from a periodic table file
    to generate the molar mass of a chemical substance in pint units.
    Since get_molar_mass returns only the value, which has inherent units of g/mol,
    this function converts to kg/mol by default, the units used for molecular weight by convention in WaterTAP.

    :param watertap_name: string name of a solute in WaterTAP format

    :return desired_quantity: molar mass of solute in pint units. Conversion from g/mol to kg/mol by default.
    """
    molar_mass_value = get_molar_mass(watertap_name)
    inherent_quantity = molar_mass_value * pyunits.g / pyunits.mol
    desired_quantity = pyunits.convert(inherent_quantity, to_units=units)
    return desired_quantity


def get_oli_names(source: dict):
    """
    Updates source dictionary with data to populate MCAS property model.

    :param source: dictionary containing WaterTAP names as keys

    :return source: dictionary with OLIName named tuples as keys
    """

    source = dict(
        map(lambda k, v: (watertap_to_oli(k), v), source.keys(), source.values())
    )
    return source


def oli_reverse_lookup(oli_name: str, names_db) -> OLIName:
    """
    Looks up WaterTAP formatted name for solute in OLI format, if listed in names_db dictionary.

    :param oli_name: string name of a solute in OLI format

    :return watertap_name: string name of a solute in WaterTAP format
    """

    if oli_name in names_db:
        return names_db[oli_name]
    else:
        raise IOError(
            f" Component {oli_name} not found in names_db."
            + " Update this dictionary to hard code additional OLI names."
        )


"""
Here follows a dictionary of OLI names and their WaterTAP counterparts.

It functions to aid reverse lookup, i.e., if a name is already in OLI format,
the necessary data can still be extracted.

TODO: method to add novel (valid) names to names_db
"""

names_db = {
    "NAION": "Na_+",
    "CLION": "Cl_-",
    "SO4ION": "SO4_2-",
    "MGION": "Mg_2+",
    "CAION": "Ca_2+",
    "KION": "K_+",
    "HCO3ION": "HCO3_-",
    "NA2CO3": "Na2CO3",
    "CO2": "CO2",
    "H2O": "H2O",
}



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for dewatering unit example.
"""

import pytest
from pyomo.environ import (
    ConcreteModel,
    value,
    assert_optimal_termination,
    units as pyunits,
)

from idaes.core import (
    FlowsheetBlock,
    MaterialBalanceType,
    MomentumBalanceType,
)

from idaes.models.unit_models.separator import SplittingType

from pyomo.environ import (
    units,
)

from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_variables,
    number_total_constraints,
    number_unused_variables,
)
import idaes.core.util.scaling as iscale
from idaes.core.util.testing import (
    initialization_tester,
)

from idaes.core.util.exceptions import (
    ConfigurationError,
)

from watertap.unit_models.dewatering import DewateringUnit, ActivatedSludgeModelType
from watertap.property_models.activated_sludge.asm1_properties import (
    ASM1ParameterBlock,
)

from watertap.property_models.activated_sludge.asm2d_properties import (
    ASM2dParameterBlock,
)
from watertap.property_models.activated_sludge.modified_asm2d_properties import (
    ModifiedASM2dParameterBlock,
)
from pyomo.util.check_units import assert_units_consistent
from watertap.costing import WaterTAPCosting
from watertap.costing.unit_models.dewatering import (
    cost_dewatering,
    cost_centrifuge,
    DewateringType,
)
from idaes.core import UnitModelCostingBlock


__author__ = "Alejandro Garciadiego, Adam Atia"

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()

# -----------------------------------------------------------------------------
@pytest.mark.unit
def test_config():
    m = ConcreteModel()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props = ASM1ParameterBlock()

    m.fs.unit = DewateringUnit(property_package=m.fs.props)

    assert len(m.fs.unit.config) == 16

    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup
    assert m.fs.unit.config.material_balance_type == MaterialBalanceType.useDefault
    assert m.fs.unit.config.momentum_balance_type == MomentumBalanceType.pressureTotal
    assert m.fs.unit.config.activated_sludge_model == ActivatedSludgeModelType.ASM1

    assert "underflow" in m.fs.unit.config.outlet_list
    assert "overflow" in m.fs.unit.config.outlet_list
    assert SplittingType.componentFlow is m.fs.unit.config.split_basis


@pytest.mark.unit
def test_list_error():
    m = ConcreteModel()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props = ASM1ParameterBlock()

    with pytest.raises(
        ConfigurationError,
        match="fs.unit encountered unrecognised "
        "outlet_list. This should not "
        "occur - please use overflow "
        "and underflow as outlets.",
    ):
        m.fs.unit = DewateringUnit(
            property_package=m.fs.props,
            outlet_list=["outlet1", "outlet2"],
        )


# -----------------------------------------------------------------------------
class TestDu(object):
    @pytest.fixture(scope="class")
    def du(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.props = ASM1ParameterBlock()

        m.fs.unit = DewateringUnit(property_package=m.fs.props)

        m.fs.unit.inlet.flow_vol.fix(178.4674 * units.m**3 / units.day)
        m.fs.unit.inlet.temperature.fix(308.15 * units.K)
        m.fs.unit.inlet.pressure.fix(1 * units.atm)

        m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(130.867 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_S"].fix(258.5789 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(
            17216.2434 * units.mg / units.liter
        )
        m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(2611.4843 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_BH"].fix(1e-6 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_BA"].fix(1e-6 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_P"].fix(626.0652 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_O"].fix(1e-6 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_NO"].fix(1e-6 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_NH"].fix(
            1442.7882 * units.mg / units.liter
        )
        m.fs.unit.inlet.conc_mass_comp[0, "S_ND"].fix(0.54323 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_ND"].fix(100.8668 * units.mg / units.liter)
        m.fs.unit.inlet.alkalinity.fix(97.8459 * units.mol / units.m**3)

        m.fs.unit.hydraulic_retention_time.fix()

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, du):
        assert hasattr(du.fs.unit, "hydraulic_retention_time")
        assert hasattr(du.fs.unit, "volume")
        assert hasattr(du.fs.unit, "inlet")
        assert len(du.fs.unit.inlet.vars) == 5
        assert hasattr(du.fs.unit.inlet, "flow_vol")
        assert hasattr(du.fs.unit.inlet, "conc_mass_comp")
        assert hasattr(du.fs.unit.inlet, "temperature")
        assert hasattr(du.fs.unit.inlet, "pressure")
        assert hasattr(du.fs.unit.inlet, "alkalinity")

        assert hasattr(du.fs.unit, "underflow")
        assert len(du.fs.unit.underflow.vars) == 5
        assert hasattr(du.fs.unit.underflow, "flow_vol")
        assert hasattr(du.fs.unit.underflow, "conc_mass_comp")
        assert hasattr(du.fs.unit.underflow, "temperature")
        assert hasattr(du.fs.unit.underflow, "pressure")
        assert hasattr(du.fs.unit.underflow, "alkalinity")

        assert hasattr(du.fs.unit, "overflow")
        assert len(du.fs.unit.overflow.vars) == 5
        assert hasattr(du.fs.unit.overflow, "flow_vol")
        assert hasattr(du.fs.unit.overflow, "conc_mass_comp")
        assert hasattr(du.fs.unit.overflow, "temperature")
        assert hasattr(du.fs.unit.overflow, "pressure")
        assert hasattr(du.fs.unit.overflow, "alkalinity")

        assert number_variables(du) == 85
        assert number_total_constraints(du) == 62
        assert number_unused_variables(du) == 6

    @pytest.mark.unit
    def test_dof(self, du):
        assert degrees_of_freedom(du) == 0

    @pytest.mark.unit
    def test_units(self, du):
        assert_units_consistent(du)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, du):
        iscale.calculate_scaling_factors(du)
        initialization_tester(du)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, du):
        solver = get_solver()
        results = solver.solve(du)
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, du):
        assert pytest.approx(101325.0, rel=1e-3) == value(
            du.fs.unit.overflow.pressure[0]
        )
        assert pytest.approx(308.15, rel=1e-3) == value(
            du.fs.unit.overflow.temperature[0]
        )
        assert pytest.approx(0.001954, rel=1e-3) == value(
            du.fs.unit.overflow.flow_vol[0]
        )
        assert pytest.approx(0.1308, rel=1e-3) == value(
            du.fs.unit.overflow.conc_mass_comp[0, "S_I"]
        )
        assert pytest.approx(0.2585, rel=1e-3) == value(
            du.fs.unit.overflow.conc_mass_comp[0, "S_S"]
        )
        assert pytest.approx(0.3638, rel=1e-3) == value(
            du.fs.unit.overflow.conc_mass_comp[0, "X_I"]
        )
        assert pytest.approx(0.0552, rel=1e-3) == value(
            du.fs.unit.overflow.conc_mass_comp[0, "X_S"]
        )
        assert value(du.fs.unit.overflow.conc_mass_comp[0, "X_BH"]) <= 1e-6
        assert value(du.fs.unit.overflow.conc_mass_comp[0, "X_BA"]) <= 1e-6
        assert pytest.approx(0.01323, rel=1e-3) == value(
            du.fs.unit.overflow.conc_mass_comp[0, "X_P"]
        )
        assert value(du.fs.unit.overflow.conc_mass_comp[0, "S_O"]) <= 1e-6
        assert value(du.fs.unit.overflow.conc_mass_comp[0, "S_NO"]) <= 1e-6
        assert pytest.approx(1.4427, rel=1e-3) == value(
            du.fs.unit.overflow.conc_mass_comp[0, "S_NH"]
        )
        assert pytest.approx(0.000543, rel=1e-3) == value(
            du.fs.unit.overflow.conc_mass_comp[0, "S_ND"]
        )
        assert pytest.approx(0.00213, rel=1e-3) == value(
            du.fs.unit.overflow.conc_mass_comp[0, "X_ND"]
        )
        assert pytest.approx(0.09784, rel=1e-3) == value(
            du.fs.unit.overflow.alkalinity[0]
        )
        assert pytest.approx(3.718, rel=1e-3) == value(du.fs.unit.volume[0])

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, du):
        assert (
            abs(
                value(
                    du.fs.unit.inlet.flow_vol[0] * du.fs.props.dens_mass
                    - du.fs.unit.overflow.flow_vol[0] * du.fs.props.dens_mass
                    - du.fs.unit.underflow.flow_vol[0] * du.fs.props.dens_mass
                )
            )
            <= 1e-6
        )
        for i in du.fs.props.solute_set:
            assert (
                abs(
                    value(
                        du.fs.unit.inlet.flow_vol[0]
                        * du.fs.unit.inlet.conc_mass_comp[0, i]
                        - du.fs.unit.overflow.flow_vol[0]
                        * du.fs.unit.overflow.conc_mass_comp[0, i]
                        - du.fs.unit.underflow.flow_vol[0]
                        * du.fs.unit.underflow.conc_mass_comp[0, i]
                    )
                )
                <= 1e-6
            )

    @pytest.mark.unit
    def test_report(self, du):
        du.fs.unit.report()


class TestDUASM2d(object):
    @pytest.fixture(scope="class")
    def du_asm2d(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.props = ASM2dParameterBlock()

        m.fs.unit = DewateringUnit(
            property_package=m.fs.props,
            activated_sludge_model=ActivatedSludgeModelType.ASM2D,
        )

        # NOTE: Concentrations of exactly 0 result in singularities, use EPS instead
        EPS = 1e-8

        m.fs.unit.inlet.flow_vol.fix(300 * units.m**3 / units.day)
        m.fs.unit.inlet.temperature.fix(308.15 * units.K)
        m.fs.unit.inlet.pressure.fix(1 * units.atm)
        m.fs.unit.inlet.conc_mass_comp[0, "S_O2"].fix(7.9707 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_N2"].fix(29.0603 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_NH4"].fix(8.0209 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_NO3"].fix(6.6395 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_PO4"].fix(7.8953 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_F"].fix(0.4748 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_A"].fix(0.0336 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(30 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(1695.7695 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(68.2975 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_H"].fix(1855.5067 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_PAO"].fix(
            214.5319 * units.mg / units.liter
        )
        m.fs.unit.inlet.conc_mass_comp[0, "X_PP"].fix(63.5316 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_PHA"].fix(2.7381 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_AUT"].fix(
            118.3582 * units.mg / units.liter
        )
        m.fs.unit.inlet.conc_mass_comp[0, "X_MeOH"].fix(EPS * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_MeP"].fix(EPS * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_TSS"].fix(
            3525.429 * units.mg / units.liter
        )
        m.fs.unit.inlet.alkalinity[0].fix(4.6663 * units.mmol / units.liter)

        m.fs.unit.hydraulic_retention_time.fix()

        return m

    @pytest.mark.unit
    def test_dof(self, du_asm2d):
        assert degrees_of_freedom(du_asm2d) == 0

    @pytest.mark.unit
    def test_units(self, du_asm2d):
        assert_units_consistent(du_asm2d)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, du_asm2d):

        iscale.calculate_scaling_factors(du_asm2d)
        initialization_tester(du_asm2d)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, du_asm2d):
        solver = get_solver()
        results = solver.solve(du_asm2d)
        assert_optimal_termination(results)


class TestDUModifiedASM2d(object):
    @pytest.fixture(scope="class")
    def du_mod_asm2d(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.props = ModifiedASM2dParameterBlock()

        m.fs.unit = DewateringUnit(
            property_package=m.fs.props,
            activated_sludge_model=ActivatedSludgeModelType.modified_ASM2D,
        )

        # NOTE: Concentrations of exactly 0 result in singularities, use EPS instead
        EPS = 1e-8

        m.fs.unit.inlet.flow_vol.fix(300 * units.m**3 / units.day)
        m.fs.unit.inlet.temperature.fix(308.15 * units.K)
        m.fs.unit.inlet.pressure.fix(1 * units.atm)
        m.fs.unit.inlet.conc_mass_comp[0, "S_O2"].fix(7.9707 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_N2"].fix(29.0603 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_NH4"].fix(8.0209 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_NO3"].fix(6.6395 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_PO4"].fix(7.8953 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_F"].fix(0.4748 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_A"].fix(0.0336 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(30 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_K"].fix(7 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_Mg"].fix(6 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "S_IC"].fix(10 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(1695.7695 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(68.2975 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_H"].fix(1855.5067 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_PAO"].fix(
            214.5319 * units.mg / units.liter
        )
        m.fs.unit.inlet.conc_mass_comp[0, "X_PP"].fix(63.5316 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_PHA"].fix(2.7381 * units.mg / units.liter)
        m.fs.unit.inlet.conc_mass_comp[0, "X_AUT"].fix(
            118.3582 * units.mg / units.liter
        )
        m.fs.unit.hydraulic_retention_time.fix()

        return m

    @pytest.mark.unit
    def test_dof(self, du_mod_asm2d):
        assert degrees_of_freedom(du_mod_asm2d) == 0

    @pytest.mark.unit
    def test_units(self, du_mod_asm2d):
        assert_units_consistent(du_mod_asm2d)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, du_mod_asm2d):

        iscale.calculate_scaling_factors(du_mod_asm2d)
        initialization_tester(du_mod_asm2d)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, du_mod_asm2d):
        solver = get_solver()
        results = solver.solve(du_mod_asm2d)
        assert_optimal_termination(results)


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_du_default_costing():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props = ASM1ParameterBlock()

    m.fs.unit = DewateringUnit(property_package=m.fs.props)

    m.fs.unit.inlet.flow_vol.fix(178.4674 * units.m**3 / units.day)
    m.fs.unit.inlet.temperature.fix(308.15 * units.K)
    m.fs.unit.inlet.pressure.fix(1 * units.atm)

    m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(130.867 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_S"].fix(258.5789 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(17216.2434 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(2611.4843 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_BH"].fix(1e-6 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_BA"].fix(1e-6 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_P"].fix(626.0652 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_O"].fix(1e-6 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NO"].fix(1e-6 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NH"].fix(1442.7882 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_ND"].fix(0.54323 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_ND"].fix(100.8668 * units.mg / units.liter)
    m.fs.unit.inlet.alkalinity.fix(97.8459 * units.mol / units.m**3)

    m.fs.unit.hydraulic_retention_time.fix()

    m.fs.costing = WaterTAPCosting()

    m.fs.unit.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
    )

    m.fs.costing.cost_process()

    assert degrees_of_freedom(m) == 0

    results = solver.solve(m)

    assert_optimal_termination(results)

    assert hasattr(m.fs.costing, "centrifuge")
    assert m.fs.unit.default_costing_method is cost_dewatering
    assert value(m.fs.costing.centrifuge.capital_a_parameter) == 328.03
    assert value(m.fs.costing.centrifuge.capital_b_parameter) == 751295

    # Check solutions
    assert pytest.approx(1964.42, rel=1e-5) == value(
        pyunits.convert(m.fs.unit.inlet.flow_vol[0], to_units=pyunits.gal / pyunits.hr)
    )
    assert pytest.approx(1602087.9 * 2, rel=1e-5) == value(
        m.fs.unit.costing.capital_cost
    )
    assert pytest.approx(1602087.9, rel=1e-5) == value(
        pyunits.convert(
            (328.03 * 1964.42 + 751295) * pyunits.USD_2007,
            to_units=m.fs.costing.base_currency,
        )
    )


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_du_centrifuge_costing():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props = ASM1ParameterBlock()

    m.fs.unit = DewateringUnit(property_package=m.fs.props)

    m.fs.unit.inlet.flow_vol.fix(178.4674 * units.m**3 / units.day)
    m.fs.unit.inlet.temperature.fix(308.15 * units.K)
    m.fs.unit.inlet.pressure.fix(1 * units.atm)

    m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(130.867 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_S"].fix(258.5789 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(17216.2434 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(2611.4843 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_BH"].fix(1e-6 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_BA"].fix(1e-6 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_P"].fix(626.0652 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_O"].fix(1e-6 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NO"].fix(1e-6 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NH"].fix(1442.7882 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_ND"].fix(0.54323 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_ND"].fix(100.8668 * units.mg / units.liter)
    m.fs.unit.inlet.alkalinity.fix(97.8459 * units.mol / units.m**3)

    m.fs.unit.hydraulic_retention_time.fix()

    m.fs.costing = WaterTAPCosting()

    m.fs.unit.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=cost_centrifuge,
    )
    # Using average specific energy consumption of 0.069 for centrifuge as a function of capacity
    m.fs.unit.energy_electric_flow_vol_inlet[0] = 0.069 * pyunits.kWh / pyunits.m**3
    m.fs.costing.cost_process()

    assert degrees_of_freedom(m) == 0

    results = solver.solve(m)

    assert_optimal_termination(results)

    assert hasattr(m.fs.costing, "centrifuge")
    assert value(m.fs.costing.centrifuge.capital_a_parameter) == 328.03
    assert value(m.fs.costing.centrifuge.capital_b_parameter) == 751295

    # Check solutions
    assert pytest.approx(1602087.9 * 2, rel=1e-5) == value(
        m.fs.unit.costing.capital_cost
    )
    assert pytest.approx(1602087.9, rel=1e-5) == value(
        pyunits.convert(
            (328.03 * 1964.42 + 751295) * pyunits.USD_2007,
            to_units=m.fs.costing.base_currency,
        )
    )
    assert pytest.approx(7.4361 * 0.069, rel=1e-5) == value(
        m.fs.unit.electricity_consumption[0]
    )


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_du_centrifuge_costing2():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props = ASM1ParameterBlock()

    m.fs.unit = DewateringUnit(property_package=m.fs.props)

    m.fs.unit.inlet.flow_vol.fix(178.4674 * units.m**3 / units.day)
    m.fs.unit.inlet.temperature.fix(308.15 * units.K)
    m.fs.unit.inlet.pressure.fix(1 * units.atm)

    m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(130.867 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_S"].fix(258.5789 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(17216.2434 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(2611.4843 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_BH"].fix(1e-6 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_BA"].fix(1e-6 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_P"].fix(626.0652 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_O"].fix(1e-6 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NO"].fix(1e-6 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NH"].fix(1442.7882 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_ND"].fix(0.54323 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_ND"].fix(100.8668 * units.mg / units.liter)
    m.fs.unit.inlet.alkalinity.fix(97.8459 * units.mol / units.m**3)

    m.fs.unit.hydraulic_retention_time.fix()

    m.fs.costing = WaterTAPCosting()

    m.fs.unit.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=cost_dewatering,
        costing_method_arguments={
            "dewatering_type": DewateringType.centrifuge,
            "cost_electricity_flow": False,
        },
    )

    m.fs.costing.cost_process()

    assert degrees_of_freedom(m) == 0

    results = solver.solve(m)

    assert_optimal_termination(results)

    assert hasattr(m.fs.costing, "centrifuge")
    assert value(m.fs.costing.centrifuge.capital_a_parameter) == 328.03
    assert value(m.fs.costing.centrifuge.capital_b_parameter) == 751295
    assert "electricity" not in m.fs.costing.aggregate_flow_costs.keys()

    # Check solutions
    assert pytest.approx(1602087.9 * 2, rel=1e-5) == value(
        m.fs.unit.costing.capital_cost
    )
    assert pytest.approx(1602087.9, rel=1e-5) == value(
        pyunits.convert(
            (328.03 * 1964.42 + 751295) * pyunits.USD_2007,
            to_units=m.fs.costing.base_currency,
        )
    )


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_du_filter_plate_press_costing():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props = ASM1ParameterBlock()

    m.fs.unit = DewateringUnit(property_package=m.fs.props)

    m.fs.unit.inlet.flow_vol.fix(178.4674 * units.m**3 / units.day)
    m.fs.unit.inlet.temperature.fix(308.15 * units.K)
    m.fs.unit.inlet.pressure.fix(1 * units.atm)

    m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(130.867 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_S"].fix(258.5789 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(17216.2434 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(2611.4843 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_BH"].fix(1e-6 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_BA"].fix(1e-6 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_P"].fix(626.0652 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_O"].fix(1e-6 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NO"].fix(1e-6 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NH"].fix(1442.7882 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_ND"].fix(0.54323 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_ND"].fix(100.8668 * units.mg / units.liter)
    m.fs.unit.inlet.alkalinity.fix(97.8459 * units.mol / units.m**3)

    m.fs.unit.hydraulic_retention_time.fix()

    m.fs.costing = WaterTAPCosting()

    m.fs.unit.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=cost_dewatering,
        costing_method_arguments={
            "dewatering_type": DewateringType.filter_plate_press,
            "cost_electricity_flow": True,
        },
    )
    # Using average specific energy consumption of 0.0039 for screw press as a function of capacity
    m.fs.unit.energy_electric_flow_vol_inlet[0] = 0.0039 * pyunits.kWh / pyunits.m**3
    m.fs.costing.cost_process()

    assert degrees_of_freedom(m) == 0

    results = solver.solve(m)

    assert_optimal_termination(results)
    assert_units_consistent(m)
    assert hasattr(m.fs.costing, "filter_plate_press")
    assert value(m.fs.costing.filter_plate_press.capital_a_parameter) == 102794
    assert value(m.fs.costing.filter_plate_press.capital_b_parameter) == 0.4216

    # Check solutions
    assert pytest.approx(2885989.2 * 2, rel=1e-5) == value(
        m.fs.unit.costing.capital_cost
    )
    assert pytest.approx(2885989.2, rel=1e-5) == value(
        pyunits.convert(
            (102794 * 1964.42**0.4216) * pyunits.USD_2007,
            to_units=m.fs.costing.base_currency,
        )
    )
    assert pytest.approx(7.4361 * 0.0039, rel=1e-5) == value(
        m.fs.unit.electricity_consumption[0]
    )


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_du_filter_belt_press_costing():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props = ASM1ParameterBlock()

    m.fs.unit = DewateringUnit(property_package=m.fs.props)

    m.fs.unit.inlet.flow_vol.fix(178.4674 * units.m**3 / units.day)
    m.fs.unit.inlet.temperature.fix(308.15 * units.K)
    m.fs.unit.inlet.pressure.fix(1 * units.atm)

    m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(130.867 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_S"].fix(258.5789 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(17216.2434 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(2611.4843 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_BH"].fix(1e-6 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_BA"].fix(1e-6 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_P"].fix(626.0652 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_O"].fix(1e-6 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NO"].fix(1e-6 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NH"].fix(1442.7882 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_ND"].fix(0.54323 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_ND"].fix(100.8668 * units.mg / units.liter)
    m.fs.unit.inlet.alkalinity.fix(97.8459 * units.mol / units.m**3)

    m.fs.unit.hydraulic_retention_time.fix()

    m.fs.costing = WaterTAPCosting()

    m.fs.unit.costing = UnitModelCostingBlock(
        flowsheet_costing_block=m.fs.costing,
        costing_method=cost_dewatering,
        costing_method_arguments={
            "dewatering_type": DewateringType.filter_belt_press,
            "cost_electricity_flow": True,
        },
    )
    # Using average specific energy consumption of 0.006 for screw press as a function of capacity
    m.fs.unit.energy_electric_flow_vol_inlet[0] = 0.006 * pyunits.kWh / pyunits.m**3
    m.fs.costing.cost_process()

    assert degrees_of_freedom(m) == 0

    results = solver.solve(m)

    assert_optimal_termination(results)
    assert_units_consistent(m)
    assert hasattr(m.fs.costing, "filter_belt_press")
    assert value(m.fs.costing.filter_belt_press.capital_a_parameter) == 146.29
    assert value(m.fs.costing.filter_belt_press.capital_b_parameter) == 433972

    # Check solutions
    assert pytest.approx(828025.2 * 2, rel=1e-5) == value(
        m.fs.unit.costing.capital_cost
    )
    assert pytest.approx(828025.2, rel=1e-5) == value(
        pyunits.convert(
            (146.29 * 1964.42 + 433972) * pyunits.USD_2007,
            to_units=m.fs.costing.base_currency,
        )
    )
    assert pytest.approx(7.4361 * 0.006, rel=1e-5) == value(
        m.fs.unit.electricity_consumption[0]
    )


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_du_costing_config_err():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props = ASM1ParameterBlock()

    m.fs.unit = DewateringUnit(property_package=m.fs.props)

    m.fs.unit.inlet.flow_vol.fix(178.4674 * units.m**3 / units.day)
    m.fs.unit.inlet.temperature.fix(308.15 * units.K)
    m.fs.unit.inlet.pressure.fix(1 * units.atm)

    m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(130.867 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_S"].fix(258.5789 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(17216.2434 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(2611.4843 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_BH"].fix(1e-6 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_BA"].fix(1e-6 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_P"].fix(626.0652 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_O"].fix(1e-6 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NO"].fix(1e-6 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NH"].fix(1442.7882 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "S_ND"].fix(0.54323 * units.mg / units.liter)
    m.fs.unit.inlet.conc_mass_comp[0, "X_ND"].fix(100.8668 * units.mg / units.liter)
    m.fs.unit.inlet.alkalinity.fix(97.8459 * units.mol / units.m**3)

    m.fs.unit.hydraulic_retention_time.fix()

    m.fs.costing = WaterTAPCosting()

    with pytest.raises(
        ConfigurationError,
        match="fs.unit received invalid argument for dewatering_type: foo. Argument must be a member of the DewateringType Enum class.",
    ):
        m.fs.unit.costing = UnitModelCostingBlock(
            flowsheet_costing_block=m.fs.costing,
            costing_method=cost_dewatering,
            costing_method_arguments={
                "dewatering_type": "foo",
            },
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of a wind-aided intensified
evaporation unit operation.
"""

from idaes.core import declare_process_block_class
from watertap.core import build_sido, constant_intensity, ZeroOrderBaseData

# Some more information about this module
__author__ = "Marcus Holly"


@declare_process_block_class("WAIVZO")
class WAIVZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a Wind-Aided Intensified Evaporation unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "waiv"

        build_sido(self)
        constant_intensity(self)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import logging
import multiprocessing
from queue import Empty as EmptyQueue

import numpy

from watertap.tools.parallel.results import LocalResults
from watertap.tools.parallel.parallel_manager import (
    parallelActor,
    ParallelManager,
)


_logger = logging.getLogger(__name__)


class MultiprocessingParallelManager(ParallelManager):
    def __init__(
        self,
        number_of_subprocesses=1,
        **kwargs,
    ):
        self.max_number_of_subprocesses = number_of_subprocesses

        # this will be updated when child processes are kicked off
        self.actual_number_of_subprocesses = None

        # Future -> (process number, parameters). Used to keep track of the process number and parameters for
        # all in-progress futures
        self.running_futures = dict()

    def is_root_process(self):
        return True

    def get_rank(self):
        return self.ROOT_PROCESS_RANK

    def number_of_worker_processes(self):
        if self.actual_number_of_subprocesses is None:
            return self.max_number_of_subprocesses
        return self.actual_number_of_subprocesses

    def sync_with_peers(self):
        pass

    def sync_array_with_peers(self, data):
        pass

    def sync_pyobject_with_peers(self, obj):
        return obj

    def combine_data_with_peers(self, data):
        return [data]

    def sum_values_and_sync(self, sendbuf, recvbuf):
        recvbuf[:] = sendbuf

    def gather_arrays_to_root(self, sendbuf, recvbuf_spec):
        receive_arr = recvbuf_spec[0]
        receive_sizes = recvbuf_spec[1]

        assert len(receive_arr) == sum(
            receive_sizes
        ), "Gathering arrays to root cannot be done with mismatched sizes"
        receive_arr[:] = sendbuf[:]

    def scatter(
        self,
        do_build,
        do_build_kwargs,
        do_execute,
        all_parameters,
    ):
        # constrain the number of child processes to the number of unique values to be run
        self.actual_number_of_subprocesses = min(
            self.max_number_of_subprocesses, len(all_parameters)
        )

        # split the parameters prameters for async run
        self.expected_samples = len(all_parameters)
        divided_parameters = numpy.array_split(all_parameters, self.expected_samples)
        # create queues, run queue will be used to store paramters we want to run
        # and return_queue is used to store results
        self.run_queue = multiprocessing.Queue()
        self.return_queue = multiprocessing.Queue()
        for i, param in enumerate(divided_parameters):
            # print(param)
            self.run_queue.put([i, param])
        # setup multiprocessing actors
        self.actors = []

        for cpu in range(self.actual_number_of_subprocesses):
            self.actors.append(
                multiprocessing.Process(
                    target=multiProcessingActor,
                    args=(
                        self.run_queue,
                        self.return_queue,
                        do_build,
                        do_build_kwargs,
                        do_execute,
                        divided_parameters[0],
                    ),
                )
            )
            self.actors[-1].start()

    def gather(self):
        results = []
        # collect result from the actors
        while len(results) < self.expected_samples:
            try:
                i, values, result = self.return_queue.get()
                results.append(LocalResults(i, values, result))
            except EmptyQueue:
                break
        self._shut_down()
        # sort the results by the process number to keep a deterministic ordering
        results.sort(key=lambda result: result.process_number)
        return results

    def _shut_down(self):
        n_shut_down = 0
        for process in self.actors:
            _logger.debug("Attempting to shut down %s", process)
            process.kill()
            n_shut_down += 1
        self.actors.clear()
        _logger.debug("Shut down %d processes", n_shut_down)

    def results_from_local_tree(self, results):
        return results


# This function is used for running the actors in multprocessing
def multiProcessingActor(
    queue,
    return_queue,
    do_build,
    do_build_kwargs,
    do_execute,
    local_parameters,
):
    actor = parallelActor(do_build, do_build_kwargs, do_execute, local_parameters)
    while True:
        try:
            msg = queue.get()
        except EmptyQueue:
            return
        i, local_parameters = msg
        result = actor.execute(local_parameters)
        return_queue.put([i, local_parameters, result])



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
from pyomo.environ import value, assert_optimal_termination
from watertap.examples.flowsheets.case_studies.municipal_treatment.municipal_treatment import (
    main,
)


# -----------------------------------------------------------------------------
@pytest.mark.component
def test_municipal_treatment():
    m, results = main()

    assert_optimal_termination(results)

    assert value(m.fs.feed.properties[0].flow_mass_comp["H2O"]) == pytest.approx(
        921.8, rel=1e-3
    )
    assert value(m.fs.feed.properties[0].flow_mass_comp["tds"]) == pytest.approx(
        0.5811, rel=1e-3
    )
    assert value(m.fs.feed.properties[0].flow_mass_comp["toc"]) == pytest.approx(
        3.690e-3, rel=1e-3
    )
    assert value(m.fs.feed.properties[0].flow_mass_comp["tss"]) == pytest.approx(
        6.019e-3, rel=1e-3
    )

    assert value(
        m.fs.backwash_pump.properties[0].flow_mass_comp["H2O"]
    ) == pytest.approx(36.87, rel=1e-3)
    assert value(
        m.fs.backwash_pump.properties[0].flow_mass_comp["tds"]
    ) == pytest.approx(0, abs=1e-6)
    assert value(
        m.fs.backwash_pump.properties[0].flow_mass_comp["toc"]
    ) == pytest.approx(0, abs=1e-6)
    assert value(
        m.fs.backwash_pump.properties[0].flow_mass_comp["tss"]
    ) == pytest.approx(5.356e-5, rel=1e-3)

    assert value(
        m.fs.recharge_pump.properties[0].flow_mass_comp["H2O"]
    ) == pytest.approx(884.8, rel=1e-3)
    assert value(
        m.fs.recharge_pump.properties[0].flow_mass_comp["tds"]
    ) == pytest.approx(5.811e-2, rel=1e-3)
    assert value(
        m.fs.recharge_pump.properties[0].flow_mass_comp["toc"]
    ) == pytest.approx(9.477e-4, rel=1e-3)
    assert value(
        m.fs.recharge_pump.properties[0].flow_mass_comp["tss"]
    ) == pytest.approx(1.656e-6, rel=1e-3)

    assert value(m.fs.costing.LCOW) == pytest.approx(4.9602e-7, rel=1e-3)  # in M$/m**3
    assert value(m.fs.costing.electricity_intensity) == pytest.approx(
        4.4812e-1, rel=1e-3
    )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
#################################################################################

import pytest

from watertap.tools.oli_api.util.watertap_to_oli_helper_functions import (
    watertap_to_oli,
    get_oli_names,
    oli_reverse_lookup,
    names_db,
    get_charge,
    get_molar_mass,
    get_molar_mass_quantity,
)
from pyomo.environ import units as pyunits, value
from pyomo.util.check_units import assert_units_equivalent

__author__ = "Paul Vecchiarelli, Adam Atia"


@pytest.mark.parametrize(
    "input_name,expected_output",
    [
        ("NaOH", "NAOH"),
        ("B[OH]3", "BOH3"),
        ("B[OH]4_-", "BOH4ION"),
        ("K_+", "KION"),
        ("Cl_-", "CLION"),
        ("Mg_2+", "MGION"),
        ("HCO3_2-", "HCO3ION"),
    ],
)
def test_watertap_to_oli(input_name: str, expected_output: str):
    assert watertap_to_oli(input_name).oli_name == expected_output


@pytest.mark.unit
def test_case_exception():
    with pytest.raises(IOError) as excinfo:
        val = "abc_2+"
        watertap_to_oli(val)
    assert (
        str(excinfo.value)
        == f" At least 1 uppercase letter is required to specify a molecule, not '{val}'."
    )


@pytest.mark.unit
def test_charge_exceptions():
    with pytest.raises(IOError) as excinfo:
        val = "Na_2#"
        watertap_to_oli(val)
    assert (
        str(excinfo.value)
        == "Only + and - are valid charge indicators and neither was provided in 'Na_2#'."
    )

    with pytest.raises(
        IOError,
        match="Charge sign could not be determined from the string 'target_ion'",
    ):
        get_charge("target_ion")
    with pytest.raises(
        IOError, match="Charge could not be determined from the string 'my_target_ion'"
    ):
        get_charge("my_target_ion")


@pytest.mark.unit
def test_get_charge():
    z = {
        "NaCl": 0,
        "Na_+": 1,
        "Cl_-": -1,
        "Ca_2+": 2,
        "SO4_2-": -2,
    }
    for solute_name, charge_value in z.items():
        assert get_charge(solute_name) == charge_value


@pytest.mark.unit
def test_get_mw():
    z = {
        "NaCl": 58.44,
        "Na_+": 22.99,
        "Cl_-": 35.45,
        "Ca_2+": 40.08,
        "SO4_2-": 96.066,
    }
    for solute_name, mw_value in z.items():
        assert get_molar_mass(solute_name) == mw_value


@pytest.mark.unit
def test_get_mw_exception():
    with pytest.raises(
        IOError, match="Molecular weight data could not be found for foo."
    ):
        get_molar_mass("foo")


@pytest.mark.unit
def test_get_mw_quantity():
    z = {
        "NaCl": 58.44,
        "Na_+": 22.99,
        "Cl_-": 35.45,
        "Ca_2+": 40.08,
        "SO4_2-": 96.066,
    }
    for solute_name, mw_value in z.items():
        result = get_molar_mass_quantity(solute_name)
        assert value(result) == mw_value / 1000
        assert_units_equivalent(result, pyunits.kg / pyunits.mol)

        result2 = get_molar_mass_quantity(solute_name, units=pyunits.g / pyunits.mol)
        assert value(result2) == mw_value
        assert_units_equivalent(result2, pyunits.g / pyunits.mol)


@pytest.mark.unit
def test_reverse_lookup():
    assert oli_reverse_lookup("NAION", names_db) == "Na_+"


@pytest.mark.unit
def test_reverse_lookup_exception():
    with pytest.raises(IOError) as excinfo:
        oli_reverse_lookup("Na_-", names_db)
    assert (
        str(excinfo.value)
        == " Component Na_- not found in names_db. Update this dictionary to hard code additional OLI names."
    )


@pytest.mark.unit
def test_get_oli_names():
    source_dict = {"Ca_2+": 500, "Cl_-": 1000, "Mg_2+": 500, "SO4_2-": 50}
    source_dict = get_oli_names(source_dict)


@pytest.mark.unit
def verify_names_db_contents():
    initial_names_db = {
        "NAION": "Na_+",
        "CLION": "Cl_-",
        "SO4ION": "SO4_2-",
        "MGION": "Mg_2+",
        "CAION": "Ca_2+",
        "KION": "K_+",
        "HCO3ION": "HCO3_-",
        "NA2CO3": "Na2CO3",
        "CO2": "CO2",
        "H2O": "H2O",
    }

    for k, v in initial_names_db.items():
        assert v == names_db[k]



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

# Import Pyomo libraries
from pyomo.environ import (
    Var,
    Suffix,
    check_optimal_termination,
    units as pyunits,
)
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (
    ControlVolume0DBlock,
    declare_process_block_class,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    UnitModelBlockData,
    useDefault,
)
from idaes.core.solvers import get_solver
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.exceptions import InitializationError
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog

from watertap.core import InitializationMixin
from watertap.costing.unit_models.compressor import cost_compressor


_log = idaeslog.getLogger(__name__)


@declare_process_block_class("Compressor")
class CompressorData(InitializationMixin, UnitModelBlockData):
    """
    Compressor model for MVC
    """

    # CONFIG are options for the unit model, this simple model only has the mandatory config options
    CONFIG = ConfigBlock()

    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
    **default** = False. The filtration unit does not support dynamic
    behavior, thus this must be False.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. The filtration unit does not have defined volume, thus
    this must be False.""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(  # TODO: update default material balance type
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
        **default** - MaterialBalanceType.useDefault.
        **Valid values:** {
        **MaterialBalanceType.useDefault - refer to property package for default
        balance type
        **MaterialBalanceType.none** - exclude material balances,
        **MaterialBalanceType.componentPhase** - use phase component balances,
        **MaterialBalanceType.componentTotal** - use total component balances,
        **MaterialBalanceType.elementTotal** - use total element balances,
        **MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.useDefault,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
        **default** - EnergyBalanceType.useDefault.
        **Valid values:** {
        **EnergyBalanceType.useDefault - refer to property package for default
        balance type
        **EnergyBalanceType.none** - exclude energy balances,
        **EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
        **EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
        **EnergyBalanceType.energyTotal** - single energy balance for material,
        **EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
        **default** - MomentumBalanceType.pressureTotal.
        **Valid values:** {
        **MomentumBalanceType.none** - exclude momentum balances,
        **MomentumBalanceType.pressureTotal** - single pressure balance for material,
        **MomentumBalanceType.pressurePhase** - pressure balances for each phase,
        **MomentumBalanceType.momentumTotal** - single momentum balance for material,
        **MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )

    def build(self):

        super().build()

        # this creates blank scaling factors, which are populated later
        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

        # Next, get the base units of measurement from the property definition
        units_meta = self.config.property_package.get_metadata().get_derived_units

        # Add unit variables
        self.pressure_ratio = Var(
            initialize=1.5, bounds=(1, None), units=pyunits.dimensionless
        )

        self.efficiency = Var(
            initialize=0.8, bounds=(0.0, 1), units=pyunits.dimensionless
        )

        # Build control volume
        self.control_volume = ControlVolume0DBlock(
            dynamic=False,
            has_holdup=False,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )

        self.control_volume.add_state_blocks(has_phase_equilibrium=False)

        self.control_volume.add_material_balances(
            balance_type=MaterialBalanceType.componentPhase, has_mass_transfer=False
        )

        self.control_volume.add_energy_balances(
            balance_type=self.config.energy_balance_type, has_work_transfer=True
        )

        self.control_volume.add_momentum_balances(
            balance_type=self.config.momentum_balance_type, has_pressure_change=True
        )

        # Add isentropic outlet block
        tmp_dict = dict(**self.config.property_package_args)
        tmp_dict["has_phase_equilibrium"] = False
        tmp_dict["parameters"] = self.config.property_package
        tmp_dict["defined_state"] = False  # isentropic outlet is not an inlet
        self.properties_isentropic_out = self.config.property_package.state_block_class(
            self.flowsheet().config.time,
            doc="Material properties of isentropic outlet",
            **tmp_dict,
        )

        # Add ports - oftentimes users interact with these rather than the state blocks
        self.add_port(name="inlet", block=self.control_volume.properties_in)
        self.add_port(name="outlet", block=self.control_volume.properties_out)

        @self.Constraint(
            self.config.property_package.phase_list,
            doc="Mass balance for inlet/isentropic outlet",
        )
        def eq_mass_balance_isentropic(b, p):
            return (
                b.control_volume.properties_in[0].flow_mass_phase_comp[p, "H2O"]
                == b.properties_isentropic_out[0].flow_mass_phase_comp[p, "H2O"]
            )

        @self.Constraint(doc="Pressure ratio")
        def eq_pressure_ratio(b):
            return (
                b.control_volume.properties_in[0].pressure * b.pressure_ratio
                == b.control_volume.properties_out[0].pressure
            )

        @self.Constraint(doc="Isentropic pressure")
        def eq_isentropic_pressure(b):
            return (
                b.properties_isentropic_out[0].pressure
                == b.control_volume.properties_out[0].pressure
            )

        @self.Constraint(doc="Isentropic temperature")
        def eq_isentropic_temperature(b):
            gamma = 1.3  # change to specific heat ratio
            return b.properties_isentropic_out[
                0
            ].temperature == b.control_volume.properties_in[
                0
            ].temperature * b.pressure_ratio ** (
                1 - 1 / gamma
            )

        @self.Constraint(doc="Energy balance/efficiency definition")
        def eq_efficiency(b):
            return (
                b.efficiency
                * (
                    b.control_volume.properties_out[0].enth_mass_phase["Vap"]
                    - b.control_volume.properties_in[0].enth_mass_phase["Vap"]
                )
                == b.properties_isentropic_out[0].enth_mass_phase["Vap"]
                - b.control_volume.properties_in[0].enth_mass_phase["Vap"]
            )

    def initialize_build(
        self, state_args=None, outlvl=idaeslog.NOTSET, solver=None, optarg=None
    ):
        """
        General wrapper for pressure changer initialization routines

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None)
            solver : str indicating which solver to use during
                     initialization (default = None)

        Returns: None
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")
        # Set solver options
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Initialize state blocks
        flags = self.control_volume.initialize(
            solver=solver, optarg=optarg, hold_state=True
        )

        init_log.info_high("Initialization Step 1 Complete.")
        # ---------------------------------------------------------------------
        # Set state_args from inlet state
        if state_args is None:
            state_args = {}
            state_dict = self.control_volume.properties_in[
                self.flowsheet().config.time.first()
            ].define_port_members()

            for k in state_dict.keys():
                if state_dict[k].is_indexed():
                    state_args[k] = {}
                    for m in state_dict[k].keys():
                        state_args[k][m] = state_dict[k][m].value
                else:
                    state_args[k] = state_dict[k].value

        self.properties_isentropic_out.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args,
        )
        init_log.info_high("Initialization Step 2 Complete.")

        # ---------------------------------------------------------------------
        # Solve unit
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 3 {}.".format(idaeslog.condition(res)))

        # ---------------------------------------------------------------------
        # Release Inlet state
        self.control_volume.release_state(flags, outlvl=outlvl)
        init_log.info("Initialization Complete: {}".format(idaeslog.condition(res)))

        if not check_optimal_termination(res):
            raise InitializationError(f"Unit model {self.name} failed to initialize")

    def _get_performance_contents(self, time_point=0):
        var_dict = {}
        var_dict["Pressure ratio"] = self.pressure_ratio
        var_dict["Efficiency"] = self.efficiency
        var_dict["Work"] = self.control_volume.work[time_point]

        return {"vars": var_dict}

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        iscale.set_scaling_factor(self.pressure_ratio, 1)
        iscale.set_scaling_factor(self.efficiency, 1)

        for j, c in self.eq_mass_balance_isentropic.items():
            sf = iscale.get_scaling_factor(
                self.control_volume.properties_in[0].flow_mass_phase_comp[j, "H2O"]
            )
            iscale.constraint_scaling_transform(c, sf)

        # Pressure constraints
        sf = iscale.get_scaling_factor(self.control_volume.properties_in[0].pressure)
        iscale.constraint_scaling_transform(self.eq_pressure_ratio, sf)
        iscale.constraint_scaling_transform(self.eq_isentropic_pressure, sf)

        # Temperature constraint
        sf = iscale.get_scaling_factor(self.control_volume.properties_in[0].temperature)
        iscale.constraint_scaling_transform(self.eq_isentropic_temperature, sf)

        # Efficiency, work constraints
        sf = iscale.get_scaling_factor(self.control_volume.work)
        iscale.constraint_scaling_transform(self.eq_efficiency, sf)

    @property
    def default_costing_method(self):
        return cost_compressor



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest
from pyomo.environ import value
from watertap.examples.flowsheets.full_treatment_train.flowsheet_components.examples import (
    full_example,
)


@pytest.mark.component
def test_flowsheet_mvp_NF_bypass_twostage_1():
    m = full_example.solve_flowsheet_mvp_NF(
        has_bypass=True,
        has_desal_feed=False,
        is_twostage=True,
        has_ERD=False,
        NF_type="ZO",
        NF_base="ion",
        RO_type="0D",
        RO_base="TDS",
        RO_level="detailed",
    )
    assert value(
        m.fs.NF.retentate.flow_mass_phase_comp[0, "Liq", "H2O"]
    ) == pytest.approx(0.1901, rel=1e-3)
    assert value(
        m.fs.NF.retentate.flow_mass_phase_comp[0, "Liq", "Ca"]
    ) == pytest.approx(1.669e-4, rel=1e-3)
    assert value(
        m.fs.RO2.retentate.flow_mass_phase_comp[0, "Liq", "H2O"]
    ) == pytest.approx(0.2408, rel=1e-3)
    assert value(
        m.fs.RO2.retentate.flow_mass_phase_comp[0, "Liq", "TDS"]
    ) == pytest.approx(2.605e-2, rel=1e-3)
    assert value(m.fs.desal_saturation.saturation_index) == pytest.approx(
        0.4082, rel=1e-3
    )


@pytest.mark.component
def test_flowsheet_mvp_cost_optimization():
    kwargs_flowsheet = {
        "has_bypass": True,
        "has_desal_feed": False,
        "is_twostage": True,
        "has_ERD": True,
        "NF_type": "ZO",
        "NF_base": "ion",
        "RO_type": "0D",
        "RO_base": "TDS",
        "RO_level": "detailed",
    }
    m = full_example.solve_optimization(system_recovery=0.68, **kwargs_flowsheet)
    assert value(
        m.fs.mixer_permeate.outlet.flow_mass_phase_comp[0, "Liq", "H2O"]
    ) == pytest.approx(0.6778, rel=1e-3)
    assert value(m.fs.costing.LCOW) == pytest.approx(0.5569, rel=1e-3)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
from watertap.ui.fsapi import FlowsheetInterface
from watertap.core.util.initialization import assert_degrees_of_freedom
from watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.supercritical_sludge_to_gas.supercritical_sludge_to_gas import (
    build,
    set_operating_conditions,
    initialize_system,
    solve,
    add_costing,
)
from pyomo.environ import units as pyunits, assert_optimal_termination
from pyomo.util.check_units import assert_units_consistent


def export_to_ui():
    return FlowsheetInterface(
        name="Supercritical sludge to gas",
        do_export=export_variables,
        do_build=build_flowsheet,
        do_solve=solve_flowsheet,
    )


def export_variables(flowsheet=None, exports=None, build_options=None, **kwargs):
    fs = flowsheet
    # --- Input data ---
    # Feed conditions
    flow_mass = (
        fs.feed.flow_mass_comp[0, "H2O"]
        + fs.feed.flow_mass_comp[0, "organic_solid"]
        + fs.feed.flow_mass_comp[0, "inorganic_solid"]
        + fs.feed.flow_mass_comp[0, "organic_liquid"]
        + fs.feed.flow_mass_comp[0, "carbon_dioxide"]
    )
    exports.add(
        obj=flow_mass,
        name="Mass flow rate",
        ui_units=pyunits.metric_ton / pyunits.day,
        display_units="ton/day",
        rounding=1,
        description="Inlet mass flow rate",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.feed.flow_mass_comp[0, "organic_solid"],
        name="Organics(s) mass flow",
        ui_units=pyunits.metric_ton / pyunits.day,
        display_units="ton/day",
        rounding=1,
        description="Inlet organics(solid) mass flow",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )
    exports.add(
        obj=fs.feed.flow_mass_comp[0, "inorganic_solid"],
        name="Inorganics(s) mass flow",
        ui_units=pyunits.metric_ton / pyunits.day,
        display_units="ton/day",
        rounding=1,
        description="Inlet inorganics(solid) mass flow",
        is_input=True,
        input_category="Feed",
        is_output=True,
        output_category="Feed",
    )

    # Unit model data, AT-HTL
    exports.add(
        obj=fs.ATHTL.recovery_frac_mass_H2O[0],
        name="Water recovery",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=2,
        description="Water recovery [g-H2O treated/g-H2O inlet]",
        is_input=True,
        input_category="AT-HTL",
        is_output=False,
    )
    exports.add(
        obj=fs.ATHTL.reaction_conversion[0, "hydrothermal_liquefaction"],
        name="Organics(s) conversion",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="Organics(s) conversion [g-organics(s) reacted/g-organics(s) inlet]",
        is_input=True,
        input_category="AT-HTL",
        is_output=False,
    )
    exports.add(
        obj=fs.ATHTL.generation_ratio["hydrothermal_liquefaction", "organic_liquid"],
        name="Organics(l) conversion ratio",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=4,
        description="Organics(l) mass conversion ratio with respect to organics(s) [g-organics(l) produced"
        "/g-organics(s) reacted]",
        is_input=True,
        input_category="AT-HTL",
        is_output=False,
    )
    exports.add(
        obj=fs.ATHTL.generation_ratio["hydrothermal_liquefaction", "carbon_dioxide"],
        name="CO2 conversion ratio",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=4,
        description="CO2 mass conversion ratio with respect to organics(s) [g-CO2 produced"
        "/g-organics(s) reacted]",
        is_input=True,
        input_category="AT-HTL",
        is_output=False,
    )
    exports.add(
        obj=fs.ATHTL.energy_electric_flow_mass,
        name="Electricity specific power",
        ui_units=pyunits.kWh / pyunits.metric_ton,
        display_units="kWh/ton of inlet flow rate",
        rounding=2,
        description="Specific energy consumption with respect to influent mass",
        is_input=True,
        input_category="AT-HTL",
        is_output=False,
    )
    exports.add(
        obj=fs.ATHTL.catalyst_dosage,
        name="Catalyst dosing",
        ui_units=pyunits.pound / pyunits.metric_ton,
        display_units="lb/ton of inlet flow rate",
        rounding=2,
        description="Catalyst dosing",
        is_input=True,
        input_category="AT-HTL",
        is_output=False,
    )

    # Unit model data, supercritical salt precipitation
    exports.add(
        obj=fs.salt_precipitation.recovery_frac_mass_H2O[0],
        name="Water recovery",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=2,
        description="Water recovery [g-H2O treated/g-H2O inlet]",
        is_input=True,
        input_category="Supercritical salt precipitation",
        is_output=False,
    )
    exports.add(
        obj=fs.salt_precipitation.removal_frac_mass_comp[0, "organic_solid"],
        name="Organics(s) removal",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=0,
        description="Organics(s) removal",
        is_input=True,
        input_category="Supercritical salt precipitation",
        is_output=False,
    )
    exports.add(
        obj=fs.salt_precipitation.removal_frac_mass_comp[0, "inorganic_solid"],
        name="Inorganics(s) removal",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=4,
        description="Inorganics(s) removal",
        is_input=True,
        input_category="Supercritical salt precipitation",
        is_output=False,
    )
    exports.add(
        obj=fs.salt_precipitation.removal_frac_mass_comp[0, "organic_liquid"],
        name="Organics(l) removal",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=0,
        description="Organics(l) removal",
        is_input=True,
        input_category="Supercritical salt precipitation",
        is_output=False,
    )
    exports.add(
        obj=fs.salt_precipitation.energy_electric_flow_mass,
        name="Electricity specific power",
        ui_units=pyunits.kWh / pyunits.metric_ton,
        display_units="kWh/ton of inlet flow rate",
        rounding=2,
        description="Specific power relating the power to inlet mass flow rate",
        is_input=True,
        input_category="Supercritical salt precipitation",
        is_output=False,
    )

    # Unit model data, HTG & WWT units
    exports.add(
        obj=fs.HTG.recovery_frac_mass_H2O[0],
        name="Water recovery",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=2,
        description="Water recovery [g-H2O treated/g-H2O inlet]",
        is_input=True,
        input_category="HTG",
        is_output=False,
    )
    exports.add(
        obj=fs.HTG.reaction_conversion[0, "hydrothermal_gasification"],
        name="Organics(l) conversion",
        ui_units=pyunits.dimensionless,
        display_units="fraction",  # we should change to %
        rounding=2,
        description="Organics(l) conversion [g-organics(l) reacted/g-organics(l) inlet]",
        is_input=True,
        input_category="HTG",
        is_output=False,
    )
    exports.add(
        obj=fs.HTG.generation_ratio["hydrothermal_gasification", "carbon_dioxide"],
        name="CO2 conversion ratio",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=4,
        description="CO2 mass conversion ratio with respect to organics(l) [g-CO2 produced"
        "/g-organics(l) reacted]",
        is_input=True,
        input_category="HTG",
        is_output=False,
    )
    exports.add(
        obj=fs.HTG.energy_electric_flow_mass,
        name="Electricity specific power",
        ui_units=pyunits.kWh / pyunits.metric_ton,
        display_units="kWh/ton of inlet flow rate",
        rounding=2,
        description="Specific power relating the power to inlet mass flow rate",
        is_input=True,
        input_category="HTG",
        is_output=False,
    )
    exports.add(
        obj=fs.HTG.catalyst_dosage,
        name="Catalyst dosing",
        ui_units=pyunits.pound / pyunits.metric_ton,
        display_units="lb/ton of inlet flow rate",
        rounding=2,
        description="Catalyst dosing",
        is_input=True,
        input_category="HTG",
        is_output=False,
    )

    # System costing
    exports.add(
        obj=fs.costing.utilization_factor,
        name="Utilization factor",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=2,
        description="Utilization factor - [annual use hours/total hours in year]",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.TIC,
        name="Practical investment factor",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=1,
        description="Practical investment factor - [total investment cost/direct "
        "capital costs]",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.plant_lifetime,
        name="Plant lifetime",
        ui_units=pyunits.year,
        display_units="years",
        rounding=1,
        description="Plant lifetime",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.wacc,
        name="Discount rate",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=2,
        description="Discount rate used in calculating the capital annualization",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.maintenance_costs_percent_FCI,
        name="Fixed operating cost factor",
        ui_units=1 / pyunits.year,
        display_units="fraction/year",
        rounding=2,
        description="Fixed operating cost factor - [annual fixed operating cost/total "
        "investment cost]",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )
    exports.add(
        obj=fs.costing.electricity_cost,
        name="Electricity cost",
        ui_units=fs.costing.base_currency / pyunits.kWh,
        display_units="$/kWh",
        rounding=3,
        description="Electricity cost",
        is_input=True,
        input_category="System costing",
        is_output=False,
    )

    # Outlets
    exports.add(
        obj=fs.product_H2O.properties[0].flow_vol,
        name="Product water flow rate",
        ui_units=pyunits.m**3 / pyunits.hr,
        display_units="m3/h",
        rounding=2,
        description="Outlet product water flow rate",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.product_H2O.properties[0].conc_mass_comp["organic_solid"],
        name="Product water organics(s) concentration",
        ui_units=pyunits.g / pyunits.L,
        display_units="g/L",
        rounding=2,
        description="Outlet product water organics(s) concentration",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.product_H2O.properties[0].conc_mass_comp["organic_liquid"],
        name="Product water organics(l) concentration",
        ui_units=pyunits.g / pyunits.L,
        display_units="g/L",
        rounding=2,
        description="Outlet product water organics(l) concentration",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )
    exports.add(
        obj=fs.product_H2O.properties[0].conc_mass_comp["inorganic_solid"],
        name="Product water inorganics(s) concentration",
        ui_units=pyunits.g / pyunits.L,
        display_units="g/L",
        rounding=2,
        description="Outlet product water inorganics(s) concentration",
        is_input=False,
        is_output=True,
        output_category="Outlets",
    )

    # System metrics
    exports.add(
        obj=fs.costing.LCOT,
        name="Levelized cost of treatment",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3 of feed",
        rounding=2,
        description="Levelized cost of treatment including operating and capital costs",
        is_input=False,
        is_output=True,
        output_category="Levelized cost metrics",
    )
    exports.add(
        obj=fs.costing.LCOW,
        name="Levelized cost of water",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3 of product",
        rounding=2,
        description="Levelized cost of water including operating and capital costs",
        is_input=False,
        is_output=True,
        output_category="Levelized cost metrics",
    )

    # Normalized metrics
    total_capital_norm = fs.costing.total_capital_cost / fs.feed.properties[0].flow_vol
    exports.add(
        obj=total_capital_norm,
        name="Total capital",
        ui_units=fs.costing.base_currency / (pyunits.m**3 / pyunits.day),
        display_units="$/(m3/day)",
        rounding=1,
        description="Normalized total capital costs accounting for indirect "
        "capital and installation - [total capital costs/feed flow rate]",
        is_input=False,
        is_output=True,
        output_category="Normalized cost metrics",
    )
    direct_capital_norm = (
        fs.ATHTL.costing.direct_capital_cost
        + fs.salt_precipitation.costing.direct_capital_cost
        + fs.HTG.costing.direct_capital_cost
    ) / fs.feed.properties[0].flow_vol
    exports.add(
        obj=direct_capital_norm,
        name="Direct capital",
        ui_units=fs.costing.base_currency / (pyunits.m**3 / pyunits.day),
        display_units="$/(m3/day)",
        rounding=1,
        description="Normalized direct capital costs - [total direct capital "
        "costs/feed flow rate] ",
        is_input=False,
        is_output=True,
        output_category="Normalized cost metrics",
    )
    elec_operating_norm = (
        fs.costing.aggregate_flow_costs["electricity"] / fs.costing.annual_water_inlet
    )
    exports.add(
        obj=elec_operating_norm,
        name="Electricity",
        ui_units=fs.costing.base_currency / pyunits.m**3,
        display_units="$/m3 of feed",
        rounding=2,
        description="Normalized electricity cost - [annual electricity costs/annual "
        "feed flow rate]",
        is_input=False,
        is_output=True,
        output_category="Normalized cost metrics",
    )

    # performance metrics
    recovery_vol = (
        fs.product_H2O.properties[0].flow_vol / fs.feed.properties[0].flow_vol
    )
    exports.add(
        obj=recovery_vol,
        name="Volumetric recovery",
        ui_units=pyunits.dimensionless,
        display_units="m3 of product/m3 of feed",
        rounding=3,
        description="Normalized volumetric recovery " "flow rate]",
        is_input=False,
        is_output=True,
        output_category="Normalized performance metrics",
    )
    removal_organics = 1 - (
        fs.product_H2O.properties[0].flow_mass_comp["organic_solid"]
        + fs.product_H2O.properties[0].flow_mass_comp["organic_liquid"]
    ) / (
        fs.feed.properties[0].flow_mass_comp["organic_solid"]
        + fs.feed.properties[0].flow_mass_comp["organic_liquid"]
    )
    exports.add(
        obj=removal_organics,
        name="Organics removal",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=3,
        description="Organics removal fraction [1 - outlet organics flow/inlet organics flow]",
        is_input=False,
        is_output=True,
        output_category="Normalized performance metrics",
    )
    removal_inorganics = (
        1
        - fs.product_H2O.properties[0].flow_mass_comp["inorganic_solid"]
        / fs.feed.properties[0].flow_mass_comp["inorganic_solid"]
    )
    exports.add(
        obj=removal_inorganics,
        name="Inorganics removal",
        ui_units=pyunits.dimensionless,
        display_units="fraction",
        rounding=3,
        description="Inorganics removal fraction [1 - outlet inorganics flow/inlet inorganics flow]",
        is_input=False,
        is_output=True,
        output_category="Normalized performance metrics",
    )

    # Capital costs
    exports.add(
        obj=fs.costing.total_capital_cost,
        name="Total",
        ui_units=fs.costing.base_currency,
        display_units="$",
        rounding=0,
        description="Total capital costs - including investment factor to account "
        "for indirect capital",
        is_input=False,
        is_output=True,
        output_category="Capital costs",
    )
    exports.add(
        obj=fs.ATHTL.costing.capital_cost,
        name="AT-HTL",
        ui_units=fs.costing.base_currency,
        display_units="$",
        rounding=0,
        description="Autothermal hydrothermal liquefaction (AT-HTL)",
        is_input=False,
        is_output=True,
        output_category="Capital costs",
    )
    exports.add(
        obj=fs.salt_precipitation.costing.capital_cost,
        name="Supercritical salt precipitation",
        ui_units=fs.costing.base_currency,
        display_units="$",
        rounding=0,
        description="Supercritical salt precipitation",
        is_input=False,
        is_output=True,
        output_category="Capital costs",
    )
    exports.add(
        obj=fs.HTG.costing.capital_cost,
        name="HTG",
        ui_units=fs.costing.base_currency,
        display_units="$",
        rounding=0,
        description="Hydrothermal gasification (HTG)",
        is_input=False,
        is_output=True,
        output_category="Capital costs",
    )

    # Operating costs
    exports.add(
        obj=fs.costing.total_operating_cost,
        name="Total",
        ui_units=fs.costing.base_currency / pyunits.year,
        display_units="$/year",
        rounding=0,
        description="Total annual operating costs - including electricity, heating, "
        "and fixed",
        is_input=False,
        is_output=True,
        output_category="Operating costs",
    )
    exports.add(
        obj=fs.costing.aggregate_flow_costs["electricity"],
        name="Electricity",
        ui_units=fs.costing.base_currency / pyunits.year,
        display_units="$/year",
        rounding=0,
        description="Annual electricity costs ",
        is_input=False,
        is_output=True,
        output_category="Operating costs",
    )
    exports.add(
        obj=fs.costing.aggregate_flow_costs["catalyst_ATHTL"],
        name="Catalyst for AT-HTL",
        ui_units=fs.costing.base_currency / pyunits.year,
        display_units="$/year",
        rounding=0,
        description="Annual catalyst cost for AT-HTL",
        is_input=False,
        is_output=True,
        output_category="Operating costs",
    )
    exports.add(
        obj=fs.costing.aggregate_flow_costs["catalyst_HTG"],
        name="Catalyst for HTG",
        ui_units=fs.costing.base_currency / pyunits.year,
        display_units="$/year",
        rounding=0,
        description="Annual catalyst cost for HTG",
        is_input=False,
        is_output=True,
        output_category="Operating costs",
    )
    exports.add(
        obj=fs.costing.total_fixed_operating_cost,
        name="Fixed",
        ui_units=fs.costing.base_currency / pyunits.year,
        display_units="$/year",
        rounding=0,
        description="Annual fixed operating costs - these costs include material "
        "replacement, maintenance, and labor",
        is_input=False,
        is_output=True,
        output_category="Operating costs",
    )


def build_flowsheet(build_options=None, **kwargs):
    # build and solve initial flowsheet
    m = build()

    set_operating_conditions(m)
    assert_degrees_of_freedom(m, 0)
    assert_units_consistent(m)

    initialize_system(m)

    results = solve(m)
    assert_optimal_termination(results)

    add_costing(m)
    assert_degrees_of_freedom(m, 0)
    m.fs.costing.initialize()

    results = solve(m)
    assert_optimal_termination(results)
    return m


def solve_flowsheet(flowsheet=None):
    fs = flowsheet
    results = solve(fs)
    return results



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest
import os
import tempfile
from watertap.examples.flowsheets.case_studies.wastewater_resource_recovery.dye_desalination.dye_sweep import (
    run_analysis,
)

# test the first 11 case studies that can be run with or without RO
pytest_parameterize_NF = list(range(1, 12))
# then test case studies 9-11 that can only run with RO
pytest_parameterize_RO = list(range(12, 18))


@pytest.mark.parametrize("case_num", pytest_parameterize_NF)
@pytest.mark.integration
def test_dye_sweep(case_num, tmp_path):
    cwd = os.getcwd()
    os.chdir(tmp_path)
    # test every other sweep with RO
    withRO = bool(case_num == 1)
    temp = tempfile.NamedTemporaryFile(delete=False)
    temp.close()
    results, params, model = run_analysis(
        case_num,
        nx=1,
        interpolate_nan_outputs=False,
        withRO=withRO,
        output_filename=temp.name,
    )
    os.remove(temp.name)
    os.chdir(cwd)
    return


@pytest.mark.parametrize("case_num", pytest_parameterize_RO)
@pytest.mark.integration
def test_dye_sweep2(case_num, tmp_path):
    cwd = os.getcwd()
    os.chdir(tmp_path)
    # test only with RO
    run_analysis(case_num, nx=1, interpolate_nan_outputs=False, withRO=True)
    os.chdir(cwd)
    return


@pytest.mark.integration
def test_withRO_condition():
    # test a case without RO for a sweep that requires RO
    with pytest.raises(Exception):
        run_analysis(15, 1, interpolate_nan_outputs=False, withRO=False)
    return


@pytest.mark.integration
def test_out_of_range_cases():
    # test a case number that is not included in the dye sweep file
    with pytest.raises(Exception):
        run_analysis(30, 1, interpolate_nan_outputs=False, withRO=False)
    return



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains the methods for constructing the material balances for
zero-order single-input/single-output (SISO) unit models (i.e. units with a single inlet and single
outlet where composition changes, such as a generic bioreactor).
"""
from types import MethodType

import idaes.logger as idaeslog
from idaes.core.solvers import get_solver
import idaes.core.util.scaling as iscale
from idaes.core.util.exceptions import InitializationError

from pyomo.environ import (
    check_optimal_termination,
    NonNegativeReals,
    Var,
    units as pyunits,
)

# Some more information about this module
__author__ = "Adam Atia"

# Set up logger
_log = idaeslog.getLogger(__name__)


def build_siso(self):
    """
    Helper method for constructing material balances for zero-order type models
    with SISO behavior.

    Two StateBlocks are added with two corresponding Ports:
        * properties_in --> inlet
        * properties_treated ---> treated


    Two additional variables are added:
        * recovery_frac_mass_H2O (indexed by time)
        * removal_frac_mass_comp (indexed by time and component)

    Two additional constraints are added to represent the material balances
        * water_recovery_equation (indexed by time)
        * solute_treated_equation (indexed by time and solute)

    This method also sets private attributes on the unit model with references
    to the appropriate initialization and scaling methods to use and to return
    the inlet volumetric flow rate.
    """
    self._has_recovery_removal = True
    self._initialize = MethodType(initialize_siso, self)
    self._scaling = MethodType(calculate_scaling_factors_siso, self)

    # Create state blocks for inlet and outlets
    tmp_dict = dict(**self.config.property_package_args)
    tmp_dict["has_phase_equilibrium"] = False
    tmp_dict["defined_state"] = True

    self.properties_in = self.config.property_package.build_state_block(
        self.flowsheet().time, doc="Material properties at inlet", **tmp_dict
    )

    tmp_dict_2 = dict(**tmp_dict)
    tmp_dict_2["defined_state"] = False

    self.properties_treated = self.config.property_package.build_state_block(
        self.flowsheet().time, doc="Material properties of treated water", **tmp_dict_2
    )

    # Create Ports
    self.add_port("inlet", self.properties_in, doc="Inlet port")
    self.add_port("treated", self.properties_treated, doc="Treated water outlet port")

    # Add performance variables
    self.recovery_frac_mass_H2O = Var(
        self.flowsheet().time,
        initialize=0.8,
        domain=NonNegativeReals,
        units=pyunits.dimensionless,
        bounds=(0.0, 1.0000001),
        doc="Mass recovery fraction of water in the treated stream",
    )
    self.removal_frac_mass_comp = Var(
        self.flowsheet().time,
        self.config.property_package.solute_set,
        domain=NonNegativeReals,
        initialize=0.01,
        units=pyunits.dimensionless,
        doc="Solute removal fraction on a mass basis",
    )

    # Add performance constraints
    # Water recovery
    @self.Constraint(self.flowsheet().time, doc="Water recovery equation")
    def water_recovery_equation(b, t):
        return (
            b.recovery_frac_mass_H2O[t] * b.properties_in[t].flow_mass_comp["H2O"]
            == b.properties_treated[t].flow_mass_comp["H2O"]
        )

    # Solute concentration of treated stream
    @self.Constraint(
        self.flowsheet().time,
        self.config.property_package.solute_set,
        doc="Constraint for solute concentration in treated " "stream.",
    )
    def solute_treated_equation(b, t, j):
        return (1 - b.removal_frac_mass_comp[t, j]) * b.properties_in[t].flow_mass_comp[
            j
        ] == b.properties_treated[t].flow_mass_comp[j]

    self._stream_table_dict = {"Inlet": self.inlet, "Treated": self.treated}

    self._perf_var_dict["Solute Removal"] = self.removal_frac_mass_comp

    self._get_Q = MethodType(_get_Q_siso, self)


def initialize_siso(
    blk, state_args=None, outlvl=idaeslog.NOTSET, solver=None, optarg=None
):
    """
    Initialization routine for single inlet-single outlet unit models.

    Keyword Arguments:
        state_args : a dict of arguments to be passed to the property
                       package(s) to provide an initial state for
                       initialization (see documentation of the specific
                       property package) (default = {}).
        outlvl : sets output level of initialization routine
        optarg : solver options dictionary object (default=None, use
                 default solver options)
        solver : str indicating which solver to use during
                 initialization (default = None, use default IDAES solver)

    Returns:
        None
    """
    if optarg is None:
        optarg = {}

    # Set solver options
    init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
    solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

    solver_obj = get_solver(solver, optarg)

    # Get initial guesses for inlet if none provided
    if state_args is None:
        state_args = {}
        state_dict = blk.properties_in[
            blk.flowsheet().time.first()
        ].define_port_members()

        for k in state_dict.keys():
            if state_dict[k].is_indexed():
                state_args[k] = {}
                for m in state_dict[k].keys():
                    state_args[k][m] = state_dict[k][m].value
            else:
                state_args[k] = state_dict[k].value

    # ---------------------------------------------------------------------
    # Initialize state blocks
    flags = blk.properties_in.initialize(
        outlvl=outlvl,
        optarg=optarg,
        solver=solver,
        state_args=state_args,
        hold_state=True,
    )
    blk.properties_treated.initialize(
        outlvl=outlvl,
        optarg=optarg,
        solver=solver,
        state_args=state_args,
        hold_state=False,
    )

    init_log.info_high("Initialization Step 1 Complete.")

    # ---------------------------------------------------------------------
    # Solve unit
    with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
        results = solver_obj.solve(blk, tee=slc.tee)

    init_log.info_high("Initialization Step 2 {}.".format(idaeslog.condition(results)))

    # ---------------------------------------------------------------------
    # Release Inlet state
    blk.properties_in.release_state(flags, outlvl)

    init_log.info("Initialization Complete: {}".format(idaeslog.condition(results)))

    if not check_optimal_termination(results):
        raise InitializationError(
            f"{blk.name} failed to initialize successfully. Please check "
            f"the output logs for more information."
        )


def calculate_scaling_factors_siso(self):
    # Get default scale factors and do calculations from base classes
    for t, v in self.water_recovery_equation.items():
        iscale.constraint_scaling_transform(
            v,
            iscale.get_scaling_factor(
                self.properties_in[t].flow_mass_comp["H2O"],
                default=1,
                warning=True,
                hint=" for water recovery",
            ),
        )

    for (t, j), v in self.solute_treated_equation.items():
        iscale.constraint_scaling_transform(
            v,
            iscale.get_scaling_factor(
                self.properties_in[t].flow_mass_comp[j], default=1, warning=False
            ),
        )  # would just be a duplicate of above


def _get_Q_siso(self, t):
    return self.properties_in[t].flow_vol



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

"""
This module contains a zero-order representation of a reactive anaerobic
digestion unit operation.
"""

from pyomo.environ import Var, units as pyunits
from idaes.core import declare_process_block_class
from watertap.core import build_sido_reactive, constant_intensity, ZeroOrderBaseData


# Some more information about this module
__author__ = "Travis Arnold"


@declare_process_block_class("AnaerobicDigestionReactiveZO")
class AnaerobicDigestionReactiveZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a reactive anaerobic digestion unit.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "anaerobic_digestion_reactive"

        build_sido_reactive(self)
        constant_intensity(self)

        if (
            self.config.process_subtype == "default"
            or self.config.process_subtype is None
        ):

            # Create water flux variable
            self.biogas_tss_ratio = Var(
                units=pyunits.m**3 / pyunits.kg,
                bounds=(0, None),
                doc="Ratio of m^3 biogas produced / kg TSS in influent",
            )
            self._perf_var_dict[
                "Ratio of m^3 biogas produced / kg TSS in influent"
            ] = self.biogas_tss_ratio
            self._fixed_perf_vars.append(self.biogas_tss_ratio)

            self.biogas_production = Var(
                self.flowsheet().time,
                units=pyunits.m**3 / pyunits.s,
                bounds=(0, None),
                doc="Biogas production",
            )
            self._perf_var_dict["Biogas production"] = self.biogas_production

            @self.Constraint(
                self.flowsheet().time, doc="Constraint for biogas production"
            )
            def biogas_prod(b, t):
                return b.biogas_production[t] == (
                    pyunits.convert(
                        b.biogas_tss_ratio * b.properties_in[t].flow_mass_comp["tss"],
                        to_units=pyunits.m**3 / pyunits.s,
                    )
                )

        if self.config.process_subtype == "GLSD_anaerobic_digester":
            self.HRT = Var(
                units=pyunits.hour,
                doc="Hydraulic retention time",
            )

            self._fixed_perf_vars.append(self.HRT)
            self._perf_var_dict["Hydraulic retention time"] = self.HRT

            self.reactor_volume = Var(
                self.flowsheet().time,
                units=pyunits.m**3,
                doc="Reactor volume",
            )

            @self.Constraint(self.flowsheet().time, doc="Constraint for reactor volume")
            def reactor_volume_cons(b, t):
                return b.reactor_volume[t] == (
                    pyunits.convert(
                        b.HRT * b.properties_in[t].flow_vol,
                        to_units=pyunits.m**3,
                    )
                )

            self._perf_var_dict["Reactor volume"] = self.reactor_volume



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pyomo.environ as pyo
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.misc import StrEnum
from ..util import cost_by_flow_volume, register_costing_parameter_block


class MixerType(StrEnum):
    default = "default"
    NaOCl = "NaOCl"
    CaOH2 = "CaOH2"


def build_mixer_cost_param_block(blk):

    blk.unit_cost = pyo.Var(
        initialize=361,
        doc="Mixer cost",
        units=pyo.units.USD_2018 / (pyo.units.liters / pyo.units.second),
    )


def cost_mixer(blk, mixer_type=MixerType.default, **kwargs):
    """
    Mixer costing method

    Args:
        mixer_type: MixerType Enum indicating mixer type,
            default = MixerType.default

        `**kwargs`: Additional keywords for the MixerType, e.g., NaOCl
            and CaOH2 mixers expect the `dosing_rate` keyword
            argument.
    """
    if mixer_type == MixerType.default:
        cost_default_mixer(blk, **kwargs)
    elif mixer_type == MixerType.NaOCl:
        cost_naocl_mixer(blk, **kwargs)
    elif mixer_type == MixerType.CaOH2:
        cost_caoh2_mixer(blk, **kwargs)
    else:
        raise ConfigurationError(
            f"{blk.unit_model.name} received invalid argument for mixer_type:"
            f" {mixer_type}. Argument must be a member of the MixerType Enum."
        )


@register_costing_parameter_block(
    build_rule=build_mixer_cost_param_block,
    parameter_block_name="mixer",
)
def cost_default_mixer(blk):
    """
    Default mixer costing method

    TODO: describe equations
    """
    cost_by_flow_volume(
        blk,
        blk.costing_package.mixer.unit_cost,
        pyo.units.convert(
            blk.unit_model.mixed_state[0].flow_vol,
            pyo.units.liter / pyo.units.second,
        ),
    )


def build_naocl_mixer_cost_param_block(blk):

    blk.unit_cost = pyo.Var(
        initialize=5.08,
        doc="NaOCl mixer cost",
        units=pyo.units.USD_2018 / (pyo.units.m**3 / pyo.units.day),
    )


def build_naocl_cost_param_block(blk):

    blk.cost = pyo.Param(
        initialize=0.23,
        doc="NaOCl cost",
        units=pyo.units.USD_2018 / pyo.units.kg,
    )
    blk.purity = pyo.Param(
        mutable=True,
        initialize=0.15,
        doc="NaOCl purity",
        units=pyo.units.dimensionless,
    )

    costing = blk.parent_block()
    costing.register_flow_type("NaOCl", blk.cost / blk.purity)


@register_costing_parameter_block(
    build_rule=build_naocl_cost_param_block, parameter_block_name="naocl"
)
@register_costing_parameter_block(
    build_rule=build_naocl_mixer_cost_param_block,
    parameter_block_name="naocl_mixer",
)
def cost_naocl_mixer(blk, dosing_rate):
    """
    NaOCl mixer costing method

    TODO: describe equations

    Args:
        dosing_rate: An expression in [mass/time] for NaOCl dosage
    """
    cost_by_flow_volume(
        blk,
        blk.costing_package.naocl_mixer.unit_cost,
        pyo.units.convert(
            blk.unit_model.inlet_stream_state[0].flow_vol,
            pyo.units.m**3 / pyo.units.day,
        ),
    )
    blk.costing_package.cost_flow(
        pyo.units.convert(dosing_rate, pyo.units.kg / pyo.units.s), "NaOCl"
    )


def build_caoh2_cost_param_block(blk):
    blk.cost = pyo.Param(
        mutable=True,
        initialize=0.12,
        doc="CaOH2 cost",
        units=pyo.units.USD_2018 / pyo.units.kg,
    )
    blk.purity = pyo.Param(
        mutable=True,
        initialize=1,
        doc="CaOH2 purity",
        units=pyo.units.dimensionless,
    )

    costing = blk.parent_block()
    costing.register_flow_type("CaOH2", blk.cost / blk.purity)


def build_caoh2_mixer_cost_param_block(blk):

    blk.unit_cost = pyo.Var(
        initialize=792.8 * 2.20462 / 2.0,
        doc="Ca(OH)2 mixer cost",
        units=pyo.units.USD_2018 / (pyo.units.kg / pyo.units.day),
    )


@register_costing_parameter_block(
    build_rule=build_caoh2_cost_param_block, parameter_block_name="caoh2"
)
@register_costing_parameter_block(
    build_rule=build_caoh2_mixer_cost_param_block,
    parameter_block_name="caoh2_mixer",
)
def cost_caoh2_mixer(blk, dosing_rate):
    """
    CaOH2 mixer costing method

    TODO: describe equations

    Args:
        dosing_rate: An expression in [mass/time] for CaOH2 dosage
    """
    stream = blk.unit_model.lime_stream
    blk.lime_kg_per_day = pyo.Expression(
        expr=pyo.units.convert(
            dosing_rate,
            pyo.units.kg / pyo.units.day,
        )
    )
    cost_by_flow_volume(
        blk,
        blk.costing_package.caoh2_mixer.unit_cost,
        blk.lime_kg_per_day,
    )
    blk.costing_package.cost_flow(
        pyo.units.convert(dosing_rate, pyo.units.kg / pyo.units.s), "CaOH2"
    )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
This module contains a zero-order representation of a METAB bioreactor with simple reactions
(i.e., conversion fractions for key reagents and conversion ratios for other reactive species).
"""

import pyomo.environ as pyo
from pyomo.environ import units as pyunits, Var
from idaes.core import declare_process_block_class
from watertap.core import build_sido_reactive, ZeroOrderBaseData

# Some more information about this module
__author__ = "Tim Bartholomew"


@declare_process_block_class("MetabZO")
class MetabZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a METAB bioreactor
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "metab"
        build_sido_reactive(self)

        self._gas_comp = self.config.process_subtype

        # unit variables
        self.volume = Var(
            initialize=1, bounds=(0, None), units=pyunits.m**3, doc="Reactor volume"
        )
        self.hydraulic_retention_time = Var(
            initialize=1,
            bounds=(0, None),
            units=pyunits.hr,
            doc="Hydraulic residence time",
        )
        self._fixed_perf_vars.append(self.hydraulic_retention_time)

        @self.Constraint(
            doc="Constraint for reactor volume based on hydraulic residence time"
        )
        def eq_reactor_volume(b):
            return b.volume == (
                pyunits.convert(
                    b.get_inlet_flow(0), to_units=pyunits.m**3 / pyunits.hour
                )
                * b.hydraulic_retention_time
            )

        # energy consumption
        self.electricity = Var(
            self.flowsheet().time,
            initialize=1,
            bounds=(0, None),
            units=pyunits.kW,
            doc="Electricity demand of unit",
        )
        self.heat = Var(
            self.flowsheet().time,
            initialize=1,
            bounds=(0, None),
            units=pyunits.kW,
            doc="Thermal demand of unit",
        )
        self.energy_electric_mixer_vol = Var(
            initialize=1,
            bounds=(0, None),
            units=pyunits.kW / pyunits.m**3,
            doc="Electricity intensity of mixer with respect to reactor volume",
        )
        self._fixed_perf_vars.append(self.energy_electric_mixer_vol)
        self.energy_electric_vacuum_flow_vol_byproduct = Var(
            initialize=1,
            bounds=(0, None),
            units=pyunits.kW / (pyunits.kg / pyunits.hr),
            doc="Electricity intensity of vacuum pump with respect to product gas flow",
        )
        self._fixed_perf_vars.append(self.energy_electric_vacuum_flow_vol_byproduct)
        self.energy_thermal_flow_vol_inlet = Var(
            initialize=1,
            bounds=(0, None),
            units=pyunits.kJ / pyunits.m**3,
            doc="Thermal energy intensity of reactor with respect to inlet volumetric flowrate",
        )
        self._fixed_perf_vars.append(self.energy_thermal_flow_vol_inlet)

        @self.Constraint(
            self.flowsheet().time,
            doc="Constraint for electricity consumption based on " "feed flowrate.",
        )
        def electricity_consumption(b, t):
            return b.electricity[t] == (
                b.energy_electric_mixer_vol * b.volume
                + b.energy_electric_vacuum_flow_vol_byproduct
                * pyunits.convert(
                    b.properties_byproduct[t].flow_mass_comp[b._gas_comp],
                    to_units=pyunits.kg / pyunits.hr,
                )
            )

        @self.Constraint(
            self.flowsheet().time,
            doc="Constraint for heat demand based on " "feed flowrate.",
        )
        def heat_demand(b, t):
            return b.heat[t] == (
                b.energy_thermal_flow_vol_inlet
                * pyunits.convert(
                    b.get_inlet_flow(t), to_units=pyunits.m**3 / pyunits.s
                )
            )

        self._perf_var_dict["Electricity Demand"] = self.electricity
        self._perf_var_dict["Thermal Energy Demand"] = self.heat

    @property
    def default_costing_method(self):
        return self.cost_metab

    @staticmethod
    def cost_metab(blk):
        """
        General method for costing the metab reactor. Capital cost
        is based on the cost of reactor, mixer, METAB beads, membrane,
        and vacuum pump.
        This method also registers the electricity demand as a costed flow.
        """
        t0 = blk.flowsheet().time.first()

        # Get parameter dict from database
        parameter_dict = blk.unit_model.config.database.get_unit_operation_parameters(
            blk.unit_model._tech_type, subtype=blk.unit_model.config.process_subtype
        )

        # Get costing parameter sub-block for this technology
        (
            reactor_cost,
            mixer_cost,
            bead_bulk_density,
            bead_cost,
            bead_replacement_factor,
            membrane_sidestream_fraction,
            membrane_specific_size,
            membrane_cost,
            vacuum_cost,
        ) = blk.unit_model._get_tech_parameters(
            blk,
            parameter_dict,
            blk.unit_model.config.process_subtype,
            [
                "reactor_cost",
                "mixer_cost",
                "bead_bulk_density",
                "bead_cost",
                "bead_replacement_factor",
                "membrane_sidestream_fraction",
                "membrane_specific_size",
                "membrane_cost",
                "vacuum_cost",
            ],
        )

        # Add capital cost variables and constraints
        blk.capital_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Capital cost of unit operation",
        )
        blk.DCC_reactor = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Direct capital cost of reactor",
        )
        blk.DCC_mixer = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Direct capital cost of mixer",
        )
        blk.DCC_bead = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Direct capital cost of beads",
        )
        blk.DCC_membrane = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Direct capital cost of membrane",
        )
        blk.DCC_vacuum = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency,
            bounds=(0, None),
            doc="Direct capital cost of vacuum pump",
        )
        blk.eq_DCC_reactor = pyo.Constraint(
            expr=blk.DCC_reactor
            == pyo.units.convert(
                blk.unit_model.volume * reactor_cost,
                to_units=blk.config.flowsheet_costing_block.base_currency,
            )
        )
        blk.eq_DCC_mixer = pyo.Constraint(
            expr=blk.DCC_mixer
            == pyo.units.convert(
                blk.unit_model.energy_electric_mixer_vol
                * blk.unit_model.volume
                * mixer_cost,
                to_units=blk.config.flowsheet_costing_block.base_currency,
            )
        )
        blk.eq_DCC_bead = pyo.Constraint(
            expr=blk.DCC_bead
            == pyo.units.convert(
                blk.unit_model.volume * bead_bulk_density * bead_cost,
                to_units=blk.config.flowsheet_costing_block.base_currency,
            )
        )
        blk.eq_DCC_membrane = pyo.Constraint(
            expr=blk.DCC_membrane
            == pyo.units.convert(
                blk.unit_model.get_inlet_flow(t0)
                * membrane_sidestream_fraction
                * membrane_specific_size
                * membrane_cost,
                to_units=blk.config.flowsheet_costing_block.base_currency,
            )
        )
        blk.eq_DCC_vacuum = pyo.Constraint(
            expr=blk.DCC_vacuum
            == pyo.units.convert(
                blk.unit_model.properties_byproduct[t0].flow_mass_comp[
                    blk.unit_model._gas_comp
                ]
                * vacuum_cost,
                to_units=blk.config.flowsheet_costing_block.base_currency,
            )
        )

        expr = (
            blk.DCC_reactor
            + blk.DCC_mixer
            + blk.DCC_bead
            + blk.DCC_membrane
            + blk.DCC_vacuum
        )

        blk.costing_package.add_cost_factor(
            blk, parameter_dict["capital_cost"]["cost_factor"]
        )

        blk.capital_cost_constraint = pyo.Constraint(
            expr=blk.capital_cost == blk.cost_factor * expr
        )

        # Add fixed operating cost variable and constraint
        blk.fixed_operating_cost = pyo.Var(
            initialize=1,
            units=blk.config.flowsheet_costing_block.base_currency
            / blk.config.flowsheet_costing_block.base_period,
            bounds=(0, None),
            doc="Fixed operating cost of unit",
        )
        blk.fixed_operating_cost_constraint = pyo.Constraint(
            expr=blk.fixed_operating_cost
            == pyo.units.convert(
                blk.DCC_bead * bead_replacement_factor,
                to_units=blk.config.flowsheet_costing_block.base_currency
                / blk.config.flowsheet_costing_block.base_period,
            )
        )

        # Register operating cost flows
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.electricity[t0], "electricity"
        )
        blk.config.flowsheet_costing_block.cost_flow(blk.unit_model.heat[t0], "heat")
        blk.config.flowsheet_costing_block.cost_flow(
            blk.unit_model.properties_byproduct[t0].flow_mass_comp[
                blk.unit_model._gas_comp
            ],
            blk.unit_model._gas_comp + "_product",
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

from idaes.core.solvers import get_solver
from watertap.tools.parameter_sweep import (
    UniformSample,
    NormalSample,
    LatinHypercubeSample,
    parameter_sweep,
)

from watertap.examples.flowsheets.RO_with_energy_recovery.RO_with_energy_recovery import (
    build,
    set_operating_conditions,
    initialize_system,
    optimize,
)

from watertap.tools.parameter_sweep import (
    get_sweep_params_from_yaml,
    set_defaults_from_yaml,
)


def get_sweep_params(m, num_samples, use_LHS=False):
    sweep_params = {}

    # Define the sampling type and ranges for three different variables
    if use_LHS:
        sweep_params["A_comp"] = LatinHypercubeSample(
            m.fs.RO.A_comp, 4.0e-12, 0.5e-12, num_samples
        )
        sweep_params["B_comp"] = LatinHypercubeSample(
            m.fs.RO.B_comp, 3.5e-8, 0.5e-8, num_samples
        )
        sweep_params["Spacer_porosity"] = LatinHypercubeSample(
            m.fs.RO.feed_side.spacer_porosity, 0.95, 0.99, num_samples
        )

    else:
        sweep_params["A_comp"] = NormalSample(
            m.fs.RO.A_comp, 4.0e-12, 0.5e-12, num_samples
        )
        sweep_params["B_comp"] = NormalSample(
            m.fs.RO.B_comp, 3.5e-8, 0.5e-8, num_samples
        )
        sweep_params["Spacer_porosity"] = UniformSample(
            m.fs.RO.feed_side.spacer_porosity, 0.95, 0.99, num_samples
        )

    return sweep_params


def build_model(
    read_model_defauls_from_file=False,
    defaults_fname="default_configuration.yaml",
):
    # Set up the solver
    solver = get_solver()

    # Build, set, and initialize the system (these steps will change depending on the underlying model)
    m = build()
    set_operating_conditions(m, water_recovery=0.5, over_pressure=0.3, solver=solver)
    initialize_system(m, solver=solver)

    # Check if we need to read in the default model values from a file
    if read_model_defauls_from_file:
        set_defaults_from_yaml(m, defaults_fname)

    return m


def build_sweep_params(
    m,
    use_LHS=False,
    sweep_params_fname="mc_sweep_params.yaml",
    read_sweep_params_from_file=False,
    num_samples=10,
):
    # Define the sampling type and ranges for three different variables
    if read_sweep_params_from_file:
        sweep_params = get_sweep_params_from_yaml(m, sweep_params_fname)
    else:
        sweep_params = get_sweep_params(m, num_samples, use_LHS=use_LHS)

    return sweep_params


def build_outputs(m):
    outputs = {}
    outputs["EC"] = m.fs.costing.specific_energy_consumption
    outputs["LCOW"] = m.fs.costing.LCOW
    return outputs


def run_parameter_sweep(
    csv_results_file_name=None,
    h5_results_file_name=None,
    seed=None,
    use_LHS=False,
    sweep_params_fname="mc_sweep_params.yaml",
    read_sweep_params_from_file=False,
    read_model_defauls_from_file=False,
    defaults_fname="default_configuration.yaml",
):
    # Set up the solver
    solver = get_solver()

    # Run the parameter sweep study using num_samples randomly drawn from the above range
    num_samples = 10

    # Run the parameter sweep
    global_results_arr, _ = parameter_sweep(
        build_model,
        build_sweep_params,
        build_outputs,
        csv_results_file_name=csv_results_file_name,
        h5_results_file_name=h5_results_file_name,
        optimize_function=optimize,
        optimize_kwargs={"solver": solver, "check_termination": False},
        num_samples=num_samples,
        seed=seed,
        build_model_kwargs=dict(
            read_model_defauls_from_file=read_model_defauls_from_file,
            defaults_fname=defaults_fname,
        ),
        build_sweep_params_kwargs=dict(
            num_samples=num_samples,
            use_LHS=use_LHS,
            sweep_params_fname=sweep_params_fname,
            read_sweep_params_from_file=read_sweep_params_from_file,
        ),
        number_of_subprocesses=1,
    )

    return global_results_arr


if __name__ == "__main__":
    # For testing this file, a seed needs to be provided as an additional argument, i.e. seed=1
    run_parameter_sweep("output/monte_carlo_results.csv")



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for zero-order centrifuge reactor model
"""
import pytest
import os

from pyomo.environ import (
    Block,
    ConcreteModel,
    Constraint,
    value,
    Var,
    assert_optimal_termination,
)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.core import UnitModelCostingBlock

from watertap.unit_models.zero_order import CentrifugeZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.costing.zero_order_costing import ZeroOrderCosting

solver = get_solver()


class TestCentrifugeZO:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(solute_list=["phosphates", "struvite"])

        m.fs.unit = CentrifugeZO(property_package=m.fs.params, database=m.db)

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(100)
        m.fs.unit.inlet.flow_mass_comp[0, "phosphates"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "struvite"].fix(0)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database is model.db
        assert model.fs.unit._tech_type == "centrifuge"

        assert isinstance(model.fs.unit.polymer_dose, Var)
        assert isinstance(model.fs.unit.polymer_demand, Var)
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)
        assert isinstance(model.fs.unit.electricity_consumption, Constraint)
        assert isinstance(model.fs.unit.polymer_demand_equation, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("centrifuge")

        model.fs.unit.load_parameters_from_database()

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert (
            model.fs.unit.recovery_frac_mass_H2O[0].value
            == data["recovery_frac_mass_H2O"]["value"]
        )

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            assert v.fixed
            assert v.value == data["removal_frac_mass_comp"][j]["value"]

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert (
            model.fs.unit.energy_electric_flow_vol_inlet.value
            == data["energy_electric_flow_vol_inlet"]["value"]
        )
        assert model.fs.unit.polymer_dose[0].fixed
        assert model.fs.unit.polymer_dose[0].value == data["polymer_dose"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(0.101, rel=1e-5) == value(
            model.fs.unit.properties_in[0].flow_vol
        )
        assert pytest.approx(9.901, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["phosphates"]
        )
        assert pytest.approx(0, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["struvite"]
        )
        assert pytest.approx(0.10005, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].flow_vol
        )
        assert pytest.approx(0.49975, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["phosphates"]
        )
        assert pytest.approx(1.503489e-09, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["struvite"]
        )
        assert pytest.approx(0.00095, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].flow_vol
        )
        assert pytest.approx(1000, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["phosphates"]
        )
        assert pytest.approx(1.583412e-7, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["struvite"]
        )
        assert pytest.approx(1.000054e-10, abs=1e-5) == value(
            model.fs.unit.electricity[0]
        )
        assert pytest.approx(174.892, rel=1e-5) == value(
            model.fs.unit.polymer_demand[0]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(
                value(
                    model.fs.unit.inlet.flow_mass_comp[0, j]
                    - model.fs.unit.treated.flow_mass_comp[0, j]
                    - model.fs.unit.byproduct.flow_mass_comp[0, j]
                )
            )

    @pytest.mark.component
    def test_report(self, model):
        model.fs.unit.report()


def test_costing():
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = WaterParameterBlock(solute_list=["phosphates", "struvite"])

    source_file = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "..",
        "..",
        "..",
        "examples",
        "flowsheets",
        "case_studies",
        "wastewater_resource_recovery",
        "amo_1575_magprex",
        "magprex_case_1575.yaml",
    )

    m.fs.costing = ZeroOrderCosting(case_study_definition=source_file)

    m.fs.unit = CentrifugeZO(property_package=m.fs.params, database=m.db)

    m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10000)
    m.fs.unit.inlet.flow_mass_comp[0, "phosphates"].fix(1)
    m.fs.unit.inlet.flow_mass_comp[0, "struvite"].fix(0)

    m.fs.unit.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit) == 0

    m.fs.unit.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    assert isinstance(m.fs.costing.centrifuge, Block)
    assert isinstance(m.fs.costing.centrifuge.HRT, Var)
    assert isinstance(m.fs.costing.centrifuge.sizing_cost, Var)

    assert isinstance(m.fs.unit.costing.capital_cost, Var)
    assert isinstance(m.fs.unit.costing.capital_cost_constraint, Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit) == 0
    initialization_tester(m)
    results = solver.solve(m)
    assert_optimal_termination(results)

    assert m.fs.unit.electricity[0] in m.fs.costing._registered_flows["electricity"]
    assert m.fs.unit.polymer_demand[0] in m.fs.costing._registered_flows["polymer"]
